# -----------------------------------------------------------------------------
# Copyright (c) 2020, 2024, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
3000 - Module for testing subscriptions
"""

import threading
import unittest

import oracledb
import test_env


class SubscriptionData:
    def __init__(self, num_messages_expected):
        self.condition = threading.Condition()
        self.num_messages_expected = num_messages_expected
        self.num_messages_received = 0

    def _process_message(self, message):
        pass

    def callback_handler(self, message):
        if message.type != oracledb.EVENT_DEREG:
            self._process_message(message)
            self.num_messages_received += 1
        if (
            message.type == oracledb.EVENT_DEREG
            or self.num_messages_received == self.num_messages_expected
        ):
            with self.condition:
                self.condition.notify()

    def wait_for_messages(self):
        if self.num_messages_received < self.num_messages_expected:
            with self.condition:
                self.condition.wait(10)


class AQSubscriptionData(SubscriptionData):
    pass


class DMLSubscriptionData(SubscriptionData):
    def __init__(self, num_messages_expected):
        super().__init__(num_messages_expected)
        self.table_operations = []
        self.row_operations = []
        self.rowids = []

    def _process_message(self, message):
        (table,) = message.tables
        self.table_operations.append(table.operation)
        for row in table.rows:
            self.row_operations.append(row.operation)
            self.rowids.append(row.rowid)


@unittest.skipIf(
    test_env.get_is_thin(), "thin mode doesn't support subscriptions"
)
class TestCase(test_env.BaseTestCase):
    @unittest.skipIf(
        test_env.get_client_version() < (23, 1), "crashes in older clients"
    )
    def test_3000(self):
        "3000 - test subscription for insert, update, delete and truncate"

        # skip if running on the Oracle Cloud, which does not support
        # subscriptions currently
        if self.is_on_oracle_cloud():
            message = "Oracle Cloud does not support subscriptions currently"
            self.skipTest(message)

        # truncate table in order to run test in known state
        self.cursor.execute("truncate table TestTempTable")

        # expected values
        table_operations = [
            oracledb.OPCODE_INSERT,
            oracledb.OPCODE_UPDATE,
            oracledb.OPCODE_INSERT,
            oracledb.OPCODE_DELETE,
            oracledb.OPCODE_ALTER | oracledb.OPCODE_ALLROWS,
        ]
        row_operations = [
            oracledb.OPCODE_INSERT,
            oracledb.OPCODE_UPDATE,
            oracledb.OPCODE_INSERT,
            oracledb.OPCODE_DELETE,
        ]
        rowids = []

        # set up subscription
        data = DMLSubscriptionData(5)
        conn = test_env.get_connection(events=True)
        sub = conn.subscribe(
            callback=data.callback_handler,
            timeout=10,
            qos=oracledb.SUBSCR_QOS_ROWIDS,
        )
        sub.registerquery("select * from TestTempTable")
        conn.autocommit = True
        cursor = conn.cursor()

        # insert statement
        cursor.execute(
            "insert into TestTempTable (IntCol, StringCol1) values (1, 'test')"
        )
        cursor.execute("select rowid from TestTempTable where IntCol = 1")
        rowids.extend(r for r, in cursor)

        # update statement
        cursor.execute(
            "update TestTempTable set StringCol1 = 'update' where IntCol = 1"
        )
        cursor.execute("select rowid from TestTempTable where IntCol = 1")
        rowids.extend(r for r, in cursor)

        # second insert statement
        cursor.execute(
            """
            insert into TestTempTable (IntCol, StringCol1)
            values (2, 'test2')
            """
        )
        cursor.execute("select rowid from TestTempTable where IntCol = 2")
        rowids.extend(r for r, in cursor)

        # delete statement
        cursor.execute("delete TestTempTable where IntCol = 2")
        rowids.append(rowids[-1])

        # truncate table
        cursor.execute("truncate table TestTempTable")

        # wait for all messages to be sent
        data.wait_for_messages()

        # verify the correct messages were sent
        self.assertEqual(data.table_operations, table_operations)
        self.assertEqual(data.row_operations, row_operations)
        self.assertEqual(data.rowids, rowids)

        # test string format of subscription object is as expected
        fmt = "<oracledb.Subscription on <oracledb.Connection to %s@%s>>"
        expected = fmt % (
            test_env.get_main_user(),
            test_env.get_connect_string(),
        )
        self.assertEqual(str(sub), expected)

    def test_3001(self):
        "3001 - test to verify deprecations"
        with self.assertRaisesFullCode("DPY-2014"):
            self.conn.subscribe(
                ip_address="www.oracle.in", ipAddress="www.oracle.in"
            )
        with self.assertRaisesFullCode("DPY-2014"):
            self.conn.subscribe(grouping_class=1, groupingClass=1)
        with self.assertRaisesFullCode("DPY-2014"):
            self.conn.subscribe(grouping_value=3, groupingValue=3)
        with self.assertRaisesFullCode("DPY-2014"):
            self.conn.subscribe(grouping_type=2, groupingType=2)
        with self.assertRaisesFullCode("DPY-2014"):
            self.conn.subscribe(client_initiated=True, clientInitiated=True)

    @unittest.skipIf(
        test_env.get_client_version() < (23, 1), "crashes in older clients"
    )
    def test_3002(self):
        "3002 - test subscription for AQ"

        # create queue and clear it of all messages
        queue = self.conn.queue("TEST_RAW_QUEUE")
        queue.deqoptions.wait = oracledb.DEQ_NO_WAIT
        while queue.deqone():
            pass
        self.conn.commit()

        # set up subscription
        data = AQSubscriptionData(1)
        conn = test_env.get_connection(events=True)
        conn.subscribe(
            namespace=oracledb.SUBSCR_NAMESPACE_AQ,
            name=queue.name,
            timeout=10,
            callback=data.callback_handler,
        )

        # enqueue a message
        queue.enqone(self.conn.msgproperties(payload="Some data"))
        self.conn.commit()

        # wait for all messages to be sent
        data.wait_for_messages()

    @unittest.skipIf(
        test_env.get_client_version() < (23, 1), "crashes in older clients"
    )
    def test_3003(self):
        "3003 - test verifying what registerquery returns"
        data = DMLSubscriptionData(5)
        qos_constants = [
            oracledb.SUBSCR_QOS_QUERY,
            oracledb.SUBSCR_QOS_RELIABLE,
            oracledb.SUBSCR_QOS_DEREG_NFY,
            oracledb.SUBSCR_QOS_ROWIDS,
            oracledb.SUBSCR_QOS_BEST_EFFORT,
        ]
        for qos_constant in qos_constants:
            conn = test_env.get_connection(events=True)
            sub = conn.subscribe(
                qos=qos_constant, callback=data.callback_handler
            )
            query_id = sub.registerquery("select * from TestTempTable")
            if qos_constant == oracledb.SUBSCR_QOS_QUERY:
                self.assertIsInstance(query_id, int)
                self.assertIsInstance(sub.id, int)
            else:
                self.assertIsNone(query_id)
            conn.unsubscribe(sub)
            conn.close()

    @unittest.skipIf(
        test_env.get_client_version() < (23, 1), "crashes in older clients"
    )
    def test_3004(self):
        "3004 - test Subscription repr()"
        data = DMLSubscriptionData(5)
        with test_env.get_connection(events=True) as conn:
            sub = conn.subscribe(callback=data.callback_handler)
            self.assertEqual(repr(sub), f"<oracledb.Subscription on {conn}>")
            conn.unsubscribe(sub)

    @unittest.skipIf(
        test_env.get_client_version() < (23, 1), "crashes in older clients"
    )
    def test_3005(self):
        "3005 - test registerquery with invalid parameters"
        data = DMLSubscriptionData(5)
        conn = test_env.get_connection(events=True)
        sub = conn.subscribe(callback=data.callback_handler)
        self.assertRaises(
            TypeError,
            sub.registerquery,
            "select * from TestTempTable",
            "invalid args",
        )
        with self.assertRaisesFullCode("ORA-00942"):
            sub.registerquery("select * from Nonexistent")
        with self.assertRaisesFullCode("DPI-1013"):
            sub.registerquery("insert into TestTempTable (IntCol) values (1)")
        conn.unsubscribe(sub)

    @unittest.skipIf(
        test_env.get_client_version() < (23, 1), "crashes in older clients"
    )
    def test_3006(self):
        "3006 - test getting subscription attributes"
        data = DMLSubscriptionData(1)
        conn = test_env.get_connection(events=True)
        cursor = conn.cursor()
        args = dict(
            callback=data.callback_handler,
            ip_address=None,
            port=0,
            name="Sub1",
            namespace=oracledb.SUBSCR_NAMESPACE_DBCHANGE,
            timeout=10,
            protocol=oracledb.SUBSCR_PROTO_OCI,
            qos=oracledb.SUBSCR_QOS_QUERY,
            operations=oracledb.OPCODE_INSERT,
        )
        sub = conn.subscribe(**args)
        for attr_name in args:
            self.assertEqual(getattr(sub, attr_name), args[attr_name])
        self.assertEqual(sub.connection, conn)
        cursor.execute("select REGID from USER_CHANGE_NOTIFICATION_REGS")
        self.assertEqual(sub.id, cursor.fetchone()[0])
        self.assertEqual(sub.ipAddress, sub.ip_address)
        conn.unsubscribe(sub)
        conn.close()

    @unittest.skipIf(
        test_env.get_client_version() < (23, 1), "crashes in older clients"
    )
    def test_3007(self):
        "3007 - test getting Message, MessageQuery, MessageTable attributes"
        condition = threading.Condition()
        conn = test_env.get_connection(events=True)

        def callback_handler(message):
            self.assertEqual(
                message.dbname.upper(), conn.instance_name.upper()
            )
            self.assertTrue(message.registered)
            self.assertEqual(message.subscription, sub)
            self.assertEqual(message.tables, [])
            self.assertIsInstance(message.txid, bytes)
            self.assertEqual(message.type, oracledb.EVENT_QUERYCHANGE)
            self.assertIsInstance(message.queries, list)
            (queries,) = message.queries
            self.assertEqual(queries.id, sub_id)
            self.assertEqual(queries.operation, oracledb.EVENT_QUERYCHANGE)
            self.assertIsInstance(queries.tables, list)
            (tables,) = queries.tables
            table_name = f"{test_env.get_main_user().upper()}.TESTTEMPTABLE"
            self.assertEqual(tables.name, table_name)
            self.assertIsInstance(tables.operation, int)
            self.assertIsInstance(tables.rows, list)
            with condition:
                condition.notify()

        sub = conn.subscribe(
            callback=callback_handler, qos=oracledb.SUBSCR_QOS_QUERY
        )
        cursor = conn.cursor()
        cursor.execute("truncate table TestTempTable")
        sub_id = sub.registerquery("select * from TestTempTable")
        cursor.execute(
            """
            insert into TestTempTable (IntCol, StringCol1)
            values (1, 'test')
            """
        )
        conn.commit()

        with condition:
            self.assertTrue(condition.wait(5))
        conn.unsubscribe(sub)

    @unittest.skipIf(
        test_env.get_client_version() < (23, 1), "crashes in older clients"
    )
    def test_3008(self):
        "3008 - test unsubscribe with invalid parameter"
        conn = test_env.get_connection(events=True)
        self.assertRaises(TypeError, conn.unsubscribe, "not a sub object")
        sub = conn.subscribe(callback=lambda x: f"Message: {x}")
        conn.unsubscribe(sub)
        with self.assertRaisesFullCode("DPI-1002"):
            conn.unsubscribe(sub)

    @unittest.skipIf(
        test_env.get_client_version() < (23, 1), "crashes in older clients"
    )
    def test_3010(self):
        "3010 - test registerquery in the middle of an active transaction"
        connection = test_env.get_connection(events=True)
        cursor = connection.cursor()
        cursor.execute("truncate table TestTempTable")
        cursor.execute(
            "insert into TestTempTable (IntCol, StringCol1) values (1, 'test')"
        )
        sub = connection.subscribe(callback=lambda x: f"Msg: {x}")
        with self.assertRaisesFullCode("ORA-29975"):
            sub.registerquery("select * from TestTempTable")
        connection.unsubscribe(sub)

    @unittest.skipIf(
        test_env.get_client_version() < (23, 1), "crashes in older clients"
    )
    def test_3011(self):
        "3011 - test registerquery with aq subscription"
        connection = test_env.get_connection(events=True)
        sub = connection.subscribe(
            callback=lambda x: None,
            namespace=oracledb.SUBSCR_NAMESPACE_AQ,
            name="TEST_RAW_QUEUE",
        )
        with self.assertRaisesFullCode("ORA-24315"):
            sub.registerquery("select * from TestTempTable")
        connection.unsubscribe(sub)

    @unittest.skipIf(
        test_env.get_client_version() < (23, 1), "crashes in older clients"
    )
    def test_3013(self):
        "3013 - test subscription with SUBSCR_QOS_DEREG_NFY deregisters"
        if self.is_on_oracle_cloud():
            self.skipTest("AQ notification not supported on the cloud")

        def callback(message):
            self.assertFalse(message.registered)
            with condition:
                condition.notify()

        condition = threading.Condition()
        self.cursor.execute("truncate table TestTempTable")
        conn = test_env.get_connection(events=True)
        cursor = conn.cursor()
        sub = conn.subscribe(
            callback=callback, qos=oracledb.SUBSCR_QOS_DEREG_NFY, timeout=2
        )
        sub.registerquery("select * from TestTempTable")
        cursor.execute(
            """
            insert into TestTempTable (IntCol, StringCol1)
            values (1, 'test')
            """
        )
        conn.commit()
        with condition:
            self.assertTrue(condition.wait(5))
        conn.unsubscribe(sub)

    @unittest.skipIf(
        test_env.get_client_version() < (23, 1), "crashes in older clients"
    )
    def test_3014(self):
        "3014 - test adding a consumer to a single consumer queue (negative)"
        conn = test_env.get_connection(events=True)
        single_consumer_queue = "TEST_RAW_QUEUE"
        with self.assertRaisesFullCode("ORA-25256"):
            conn.subscribe(
                callback=lambda x: None,
                namespace=oracledb.SUBSCR_NAMESPACE_AQ,
                name=f"{single_consumer_queue}:SUBSCRIBER",
            )


if __name__ == "__main__":
    test_env.run_test_cases()



# -----------------------------------------------------------------------------
# Copyright (c) 2023, 2024, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
6700 - Module for testing the JSON data type extension in Oracle Database 23ai.
"""

import json
import unittest

import oracledb
import test_env


@unittest.skipUnless(
    test_env.get_client_version() >= (23, 0), "unsupported client"
)
@unittest.skipUnless(
    test_env.get_server_version() >= (23, 0), "unsupported server"
)
class TestCase(test_env.BaseTestCase):
    def __test_fetch_json(self, value, table_name="TestJson"):
        """
        Tests fetching JSON encoded by the database.
        """
        self.cursor.execute(f"delete from {table_name}")
        self.cursor.execute(
            f"insert into {table_name} values (1, :1)", [json.dumps(value)]
        )
        self.cursor.execute(f"select JsonCol from {table_name}")
        (fetched_value,) = self.cursor.fetchone()
        self.assertEqual(fetched_value, value)

    def __test_round_trip_json(self, value):
        """
        Tests fetching JSON encoded by the driver.
        """
        self.cursor.execute("delete from TestJson")
        self.cursor.setinputsizes(oracledb.DB_TYPE_JSON)
        self.cursor.execute("insert into TestJson values (1, :1)", [value])
        self.cursor.execute("select JsonCol from TestJson")
        (fetched_value,) = self.cursor.fetchone()
        self.assertEqual(fetched_value, value)

    def test_6700(self):
        "6700 - fetch JSON with a field name greater than 255 bytes"
        fname_long = "A" * 256
        value = {}
        value[fname_long] = 6700
        self.__test_fetch_json(value)

    def test_6701(self):
        "6701 - fetch JSON with field names greater and less than 255 bytes"
        fname_short = "short_name"
        fname_long = "A" * 256
        value = {}
        value[fname_short] = "Short name"
        value[fname_long] = 6701
        self.__test_fetch_json(value)

    def test_6702(self):
        "6702 - fetch JSON with many field names greater than 255 bytes"
        value = {}
        for i in range(26):
            for j in range(26):
                fname = chr(i + ord("A")) + chr(j + ord("A")) + "X" * 254
                value[fname] = 12.25
        self.__test_fetch_json(value)

    def test_6703(self):
        "6703 - fetch JSON with many field names (large and small)"
        value = {}
        for i in range(26):
            for j in range(26):
                short_name = chr(i + ord("A")) + chr(j + ord("A"))
                value[short_name] = 6.75
                long_name = short_name + "X" * 254
                value[long_name] = 12.25
        self.__test_fetch_json(value)

    def test_6704(self):
        "6704 - fetch JSON with many field names (one large and many small)"
        value = {}
        long_name = "B" * 256
        value[long_name] = 6704
        for i in range(26):
            for j in range(26):
                short_name = chr(i + ord("A")) + chr(j + ord("A"))
                value[short_name] = 8.625
        self.__test_fetch_json(value)

    def test_6705(self):
        "6705 - round trip JSON with a field name greater than 255 bytes"
        fname_long = "A" * 256
        value = {}
        value[fname_long] = 6705
        self.__test_round_trip_json(value)

    def test_6706(self):
        "6706 - round trip JSON with field names (small and large)"
        fname_short = "short_name"
        fname_long = "A" * 256
        value = {}
        value[fname_short] = "Short name"
        value[fname_long] = 6706
        self.__test_round_trip_json(value)

    def test_6707(self):
        "6707 - round trip JSON with many field names greater than 255 bytes"
        value = {}
        for i in range(26):
            for j in range(26):
                fname = chr(i + ord("A")) + chr(j + ord("A")) + "X" * 254
                value[fname] = 12.25
        self.__test_round_trip_json(value)

    def test_6708(self):
        "6708 - round trip JSON with many field names (large and small)"
        value = {}
        for i in range(26):
            for j in range(26):
                short_name = chr(i + ord("A")) + chr(j + ord("A"))
                value[short_name] = 6.75
                long_name = short_name + "X" * 254
                value[long_name] = 12.25
        self.__test_round_trip_json(value)

    def test_6709(self):
        "6709 - round trip JSON with many field names (1 large and many small)"
        value = {}
        long_name = "B" * 256
        value[long_name] = 6704
        for i in range(26):
            for j in range(26):
                short_name = chr(i + ord("A")) + chr(j + ord("A"))
                value[short_name] = 8.625
        self.__test_round_trip_json(value)

    def test_6710(self):
        "6710 - fetch JSON with relative offsets"
        value = {}
        fname_long = "C" * 256
        value[fname_long] = 6710
        value["num_list"] = [1.5, 2.25, 3.75, 5.5]
        value["str_list"] = ["string 1", "string 2"]
        self.__test_fetch_json(value, "TestCompressedJson")

    def test_6711(self):
        "6711 - fetch JSON with relative offsets and shared fields and values"
        value = []
        for i in range(15):
            value.append(dict(a=6711, b="String Value"))
        self.__test_fetch_json(value, "TestCompressedJson")

    def test_6712(self):
        "6712 - fetch JSON with relative offsets and shared fields, not values"
        value = []
        for i in range(15):
            value.append(dict(a=6711 + i, b=f"String Value {i}"))
        self.__test_fetch_json(value, "TestCompressedJson")


if __name__ == "__main__":
    test_env.run_test_cases()



#------------------------------------------------------------------------------
# Copyright (c) 2020, 2025, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# connect.pyx
#
# Cython file defining the messages that are sent to the database and the
# responses that are received by the client for the initial connection request
# (embedded in thin_impl.pyx).
#------------------------------------------------------------------------------

@cython.final
cdef class ConnectMessage(Message):
    cdef:
        bytes connect_string_bytes
        uint16_t connect_string_len, redirect_data_len
        bint read_redirect_data_len
        Description description
        uint8_t packet_flags
        str redirect_data
        str host
        int port

    cdef int process(self, ReadBuffer buf) except -1:
        cdef:
            uint16_t protocol_version, protocol_options
            const char_type *redirect_data
            uint32_t flags2 = 0
            uint8_t flags1
            bytes db_uuid
        if buf._current_packet.packet_type == TNS_PACKET_TYPE_REDIRECT:
            if not self.read_redirect_data_len:
                buf.read_uint16be(&self.redirect_data_len)
                self.read_redirect_data_len = True
            buf.wait_for_packets_sync()
            redirect_data = buf.read_raw_bytes(self.redirect_data_len)
            if self.redirect_data_len > 0:
                self.redirect_data = \
                        redirect_data[:self.redirect_data_len].decode()
            self.read_redirect_data_len = False
        elif buf._current_packet.packet_type == TNS_PACKET_TYPE_ACCEPT:
            buf.read_uint16be(&protocol_version)
            # check if the protocol version supported by the database is high
            # enough; if not, reject the connection immediately
            if protocol_version < TNS_VERSION_MIN_ACCEPTED:
                errors._raise_err(errors.ERR_SERVER_VERSION_NOT_SUPPORTED)
            buf.read_uint16be(&protocol_options)
            buf.skip_raw_bytes(10)
            buf.read_ub1(&flags1)
            if flags1 & TNS_NSI_NA_REQUIRED:
                feature = "Native Network Encryption and Data Integrity"
                errors._raise_not_supported(feature)
            buf.skip_raw_bytes(9)
            buf.read_uint32be(&buf._caps.sdu)
            if protocol_version >= TNS_VERSION_MIN_OOB_CHECK:
                buf.skip_raw_bytes(5)
                buf.read_uint32be(&flags2)
            buf._caps._adjust_for_protocol(protocol_version, protocol_options,
                                           flags2)
            buf._transport._full_packet_size = True
        elif buf._current_packet.packet_type == TNS_PACKET_TYPE_REFUSE:
            response = self.error_info.message
            error_code = "unknown"
            error_code_int = 0
            if response is not None:
                pos = response.find("(ERR=")
                if pos > 0:
                    end_pos = response.find(")", pos)
                    if end_pos > 0:
                        error_code = response[pos + 5:end_pos]
                        error_code_int = int(error_code)
            if error_code_int == 0:
                errors._raise_err(errors.ERR_UNEXPECTED_REFUSE)
            if error_code_int == TNS_ERR_INVALID_SERVICE_NAME:
                errors._raise_err(errors.ERR_INVALID_SERVICE_NAME,
                                  service_name=self.description.service_name,
                                  host=self.host, port=self.port)
            elif error_code_int == TNS_ERR_INVALID_SID:
                errors._raise_err(errors.ERR_INVALID_SID,
                                  sid=self.description.sid,
                                  host=self.host, port=self.port)
            errors._raise_err(errors.ERR_LISTENER_REFUSED_CONNECTION,
                              error_code=error_code)

    cdef int send(self, WriteBuffer buf) except -1:
        cdef:
            uint16_t service_options = TNS_GSO_DONT_CARE
            uint32_t connect_flags_1 = 0, connect_flags_2 = 0
            uint8_t nsi_flags = \
                    TNS_NSI_SUPPORT_SECURITY_RENEG | TNS_NSI_DISABLE_NA
        if buf._caps.supports_oob:
            service_options |= TNS_GSO_CAN_RECV_ATTENTION
            connect_flags_2 |= TNS_CHECK_OOB
        buf.start_request(TNS_PACKET_TYPE_CONNECT, self.packet_flags)
        buf.write_uint16be(TNS_VERSION_DESIRED)
        buf.write_uint16be(TNS_VERSION_MINIMUM)
        buf.write_uint16be(service_options)
        buf.write_uint16be(self.description.sdu)
        buf.write_uint16be(self.description.sdu)
        buf.write_uint16be(TNS_PROTOCOL_CHARACTERISTICS)
        buf.write_uint16be(0)               # line turnaround
        buf.write_uint16be(1)               # value of 1
        buf.write_uint16be(self.connect_string_len)
        buf.write_uint16be(74)              # offset to connect data
        buf.write_uint32be(0)               # max receivable data
        buf.write_uint8(nsi_flags)
        buf.write_uint8(nsi_flags)
        buf.write_uint64be(0)               # obsolete
        buf.write_uint64be(0)               # obsolete
        buf.write_uint64be(0)               # obsolete
        buf.write_uint32be(self.description.sdu)      # SDU (large)
        buf.write_uint32be(self.description.sdu)      # TDU (large)
        buf.write_uint32be(connect_flags_1)
        buf.write_uint32be(connect_flags_2)
        if self.connect_string_len > TNS_MAX_CONNECT_DATA:
            buf.end_request()
            buf.start_request(TNS_PACKET_TYPE_DATA)
        buf.write_bytes(self.connect_string_bytes)
        buf.end_request()



# -----------------------------------------------------------------------------
# Copyright (c) 2023, 2024, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
6400 - Module for testing the VECTOR database type
"""

import array
import json
import unittest
import oracledb
import test_env


@unittest.skipUnless(
    test_env.get_client_version() >= (23, 4), "unsupported client"
)
@unittest.skipUnless(
    test_env.get_server_version() >= (23, 4), "unsupported server"
)
class TestCase(test_env.BaseTestCase):
    def __test_insert_and_fetch(self, value, column_name, expected_typecode):
        """
        Test inserting and fetching a vector.
        """
        self.cursor.execute("delete from TestVectors")
        if isinstance(value, list):
            self.cursor.setinputsizes(value=oracledb.DB_TYPE_VECTOR)
        self.cursor.execute(
            f"""
            insert into TestVectors (IntCol, {column_name})
            values(1, :value)
            """,
            value=value,
        )
        self.conn.commit()
        self.cursor.execute(f"select {column_name} from TestVectors")
        (fetched_value,) = self.cursor.fetchone()
        if expected_typecode == "b":
            expected_value = array.array("b", [int(i) for i in value])
        else:
            expected_value = array.array(expected_typecode, value)
        self.assertEqual(fetched_value, expected_value)
        self.assertEqual(fetched_value.typecode, expected_typecode)

    def __test_plsql_insert_and_fetch(self, vec1, vec2, expected_distance):
        in_out_vec = self.cursor.var(oracledb.DB_TYPE_VECTOR)
        in_out_vec.setvalue(0, vec2)

        distance = self.cursor.var(oracledb.DB_TYPE_BINARY_DOUBLE)
        output_vec = self.cursor.var(oracledb.DB_TYPE_VECTOR)

        plsql_block = """
            BEGIN
                select
                    vector_distance(:in_vec, :in_out_vec, euclidean)
                    into :distance;
                :output_vec := :in_out_vec;
                :in_out_vec := :in_vec;
            END;
            """

        self.cursor.execute(
            plsql_block,
            in_vec=vec1,
            in_out_vec=in_out_vec,
            distance=distance,
            output_vec=output_vec,
        )
        self.assertEqual(output_vec.getvalue(), vec2)
        self.assertEqual(in_out_vec.getvalue(), vec1)
        self.assertAlmostEqual(
            distance.getvalue(), expected_distance, places=2
        )

    def test_6400(self):
        "6400 - test binding in a vector from a Python list"
        value = [1, 2]
        self.cursor.setinputsizes(oracledb.DB_TYPE_VECTOR)
        self.cursor.execute("select :1 from dual", [value])
        (fetched_value,) = self.cursor.fetchone()
        self.assertIsInstance(fetched_value, array.array)
        self.assertEqual(fetched_value.typecode, "d")
        self.assertEqual(fetched_value, array.array("d", value))

    def test_6401(self):
        "6401 - test binding in a vector from a Python array of type float64"
        value = array.array("d", [3, 4, 5])
        self.cursor.execute("select :1 from dual", [value])
        (fetched_value,) = self.cursor.fetchone()
        self.assertIsInstance(fetched_value, array.array)
        self.assertEqual(fetched_value.typecode, "d")
        self.assertEqual(fetched_value, value)

    def test_6402(self):
        "6402 - test binding in a vector from a Python array of type float32"
        value = array.array("f", [6, 7, 8, 9])
        self.cursor.execute("select :1 from dual", [value])
        (fetched_value,) = self.cursor.fetchone()
        self.assertIsInstance(fetched_value, array.array)
        self.assertEqual(fetched_value.typecode, "f")
        self.assertEqual(fetched_value, value)

    def test_6403(self):
        "6402 - test binding in a vector from a Python array of type int8"
        value = array.array("b", [-10, 11, -12, 13, -14])
        self.cursor.execute("select :1 from dual", [value])
        (fetched_value,) = self.cursor.fetchone()
        self.assertIsInstance(fetched_value, array.array)
        self.assertEqual(fetched_value.typecode, "b")
        self.assertEqual(fetched_value, value)

    def test_6404(self):
        "6404 - unspported array type for vector"
        with self.assertRaisesFullCode("DPY-3013"):
            self.cursor.execute(
                "select :1 from dual", [array.array("L", [4, 5])]
            )

    def test_6405(self):
        "6405 - insert a float32 vector into a float32 column"
        value = array.array(
            "f",
            [
                1.23,
                4.56,
                -7.89,
                10.11,
                -12.13,
                14.15,
                -16.17,
                18.19,
                -20.21,
                9.23,
                -2.54,
                6.5,
                4.21,
                -1.96,
                3.54,
                2.6,
            ],
        )
        self.__test_insert_and_fetch(value, "Vector32Col", "f")

    def test_6406(self):
        "6406 - insert a float32 vector into a float64 column"
        value = array.array(
            "f",
            [
                1.23,
                4.56,
                -7.89,
                10.11,
                -12.13,
                14.15,
                -16.17,
                18.19,
                -20.21,
                9.23,
                -2.54,
                6.5,
                4.21,
                -1.96,
                3.54,
                2.6,
            ],
        )
        self.__test_insert_and_fetch(value, "Vector64Col", "d")

    def test_6407(self):
        "6407 - insert a float32 vector into a flexible format column"
        value = array.array(
            "f",
            [
                1.23,
                4.56,
                -7.89,
                10.11,
                -12.13,
                14.15,
                -16.17,
                18.19,
                -20.21,
                9.23,
                -2.54,
                6.5,
                4.21,
                -1.96,
                3.54,
                2.6,
            ],
        )
        self.__test_insert_and_fetch(value, "VectorFlexAllCol", "f")

    def test_6408(self):
        "6408 - insert a float64 vector into a float64 column"
        value = array.array(
            "d",
            [
                -0.0375,
                0.625,
                -0.025,
                0.125,
                -0.75,
                0.0,
                -0.3625,
                0.125,
                -0.5,
                0.03125,
                -2.50,
                -0.75,
                1.625,
                1.025,
                0.125,
                0.725,
            ],
        )
        self.__test_insert_and_fetch(value, "Vector64Col", "d")

    def test_6409(self):
        "6409 - insert float64 vector into a float32 column"
        value = array.array(
            "d",
            [
                -0.0375,
                0.625,
                -0.025,
                0.125,
                -0.75,
                0.0,
                -0.3625,
                0.125,
                -0.5,
                0.03125,
                -2.50,
                -0.75,
                1.625,
                1.025,
                0.125,
                0.725,
            ],
        )
        self.__test_insert_and_fetch(value, "Vector32Col", "f")

    def test_6410(self):
        "6410 - insert float64 vector into a flexible type column"
        value = array.array(
            "d",
            [
                -0.0375,
                0.625,
                -0.025,
                0.125,
                -0.75,
                0.0,
                -0.3625,
                0.125,
                -0.5,
                0.03125,
                -2.50,
                -0.75,
                1.625,
                1.025,
                0.125,
                0.725,
            ],
        )
        self.__test_insert_and_fetch(value, "VectorFlexAllCol", "d")

    def test_6411(self):
        "6411 - insert a vector with an invalid size"
        self.cursor.execute("truncate table TestVectors")
        for num_elems in [4, 20]:
            statement = """
                    insert into TestVectors (IntCol, Vector64Col)
                    values(2, :1)"""
            vector = array.array("d", [i * 0.625 for i in range(num_elems)])
            with self.assertRaisesFullCode("ORA-51803"):
                self.cursor.execute(statement, [vector])

    def test_6412(self):
        "6412 - verify fetch info for vectors"
        attr_names = [
            "name",
            "type_code",
            "vector_dimensions",
            "vector_format",
            "vector_is_sparse",
        ]
        expected_values = [
            ["INTCOL", oracledb.DB_TYPE_NUMBER, None, None, None],
            ["VECTORFLEXALLCOL", oracledb.DB_TYPE_VECTOR, None, None, False],
            ["VECTORFLEXTYPECOL", oracledb.DB_TYPE_VECTOR, 2, None, False],
            [
                "VECTORFLEX8COL",
                oracledb.DB_TYPE_VECTOR,
                None,
                oracledb.VECTOR_FORMAT_INT8,
                False,
            ],
            [
                "VECTORFLEX32COL",
                oracledb.DB_TYPE_VECTOR,
                None,
                oracledb.VECTOR_FORMAT_FLOAT32,
                False,
            ],
            [
                "VECTORFLEX64COL",
                oracledb.DB_TYPE_VECTOR,
                None,
                oracledb.VECTOR_FORMAT_FLOAT64,
                False,
            ],
            [
                "VECTOR8COL",
                oracledb.DB_TYPE_VECTOR,
                16,
                oracledb.VECTOR_FORMAT_INT8,
                False,
            ],
            [
                "VECTOR32COL",
                oracledb.DB_TYPE_VECTOR,
                16,
                oracledb.VECTOR_FORMAT_FLOAT32,
                False,
            ],
            [
                "VECTOR64COL",
                oracledb.DB_TYPE_VECTOR,
                16,
                oracledb.VECTOR_FORMAT_FLOAT64,
                False,
            ],
        ]
        self.cursor.execute("select * from TestVectors")
        values = [
            [getattr(i, n) for n in attr_names]
            for i in self.cursor.description
        ]
        self.assertEqual(values, expected_values)
        self.assertIs(
            self.cursor.description[3].vector_format,
            oracledb.VectorFormat.INT8,
        )

    def test_6413(self):
        "6413 - insert an int8 vector into an int8 column"
        value = array.array(
            "b", [-5, 4, -7, 6, -9, 8, -127, 127, 0, -128, 1, 4, -3, 2, -8, 0]
        )
        self.__test_insert_and_fetch(value, "Vector8Col", "b")

    def test_6414(self):
        "6414 - insert an int8 vector into a float32 column"
        value = array.array(
            "b", [-5, 4, -7, 6, -9, 8, -127, 127, 0, -128, 1, 4, -3, 2, -8, 0]
        )
        self.__test_insert_and_fetch(value, "Vector32Col", "f")

    def test_6415(self):
        "6415 - insert an int8 vector into a float64 column"
        value = array.array(
            "b", [-5, 4, -7, 6, -9, 8, -127, 127, 0, -128, 1, 4, -3, 2, -8, 0]
        )
        self.__test_insert_and_fetch(value, "Vector64Col", "d")

    def test_6416(self):
        "6416 - insert an int8 vector into a flexible column"
        value = array.array(
            "b", [-5, 4, -7, 6, -9, 8, -127, 127, 0, -128, 1, 4, -3, 2, -8, 0]
        )
        self.__test_insert_and_fetch(value, "VectorFlexAllCol", "b")

    def test_6417(self):
        "6417 - insert a float32 vector into an int8 column"
        value = array.array(
            "f", [-5, 4, -7, 6, -9, 8, -127, 127, 0, -128, 1, 4, -3, 2, -8, 0]
        )
        self.__test_insert_and_fetch(value, "Vector8Col", "b")

    def test_6418(self):
        "6418 - insert a float64 vector into an int8 column"
        value = array.array(
            "d", [-5, 4, -7, 6, -9, 8, -127, 127, 0, -128, 1, 4, -3, 2, -8, 0]
        )
        self.__test_insert_and_fetch(value, "Vector8Col", "b")

    def test_6419(self):
        "6419 - test dml returning vector type"
        value = array.array("d", [6423.5, 6423.625])
        out_var = self.cursor.var(oracledb.DB_TYPE_VECTOR)
        self.cursor.execute("delete from TestVectors")
        self.cursor.execute(
            """
            insert into TestVectors (IntCol, VectorFlexTypeCol)
            values (1, :value)
            returning VectorFlexTypeCol into :out_value
            """,
            [value, out_var],
        )
        self.conn.commit()
        self.assertEqual(value, out_var.getvalue()[0])

    def test_6420(self):
        "6420 - test handling of NULL vector value"
        self.cursor.execute("delete from TestVectors")
        self.cursor.execute("insert into TestVectors (IntCol) values (1)")
        self.conn.commit()
        self.cursor.execute("select VectorFlexTypeCol from TestVectors")
        (fetched_value,) = self.cursor.fetchone()
        self.assertIsNone(fetched_value)

    def test_6421(self):
        "6421 - insert a float32 vector into an int8 column (negative)"
        value = array.array(
            "f",
            [-130, -129, 0, 1, 2, 3, 127, 128, 129, 348, 12, 49, 78, 12, 9, 2],
        )
        with self.assertRaisesFullCode("ORA-51806"):
            self.__test_insert_and_fetch(value, "Vector8Col", "b")

    def test_6422(self):
        "6422 - insert a float64 vector with 65,533 dimensions"
        value = array.array("d", [2.5] * 65533)
        self.__test_insert_and_fetch(value, "VectorFlexAllCol", "d")

    def test_6423(self):
        "6423 - insert a float32 vector with 65,533 dimensions"
        value = array.array("f", [2.5] * 65533)
        self.__test_insert_and_fetch(value, "VectorFlexAllCol", "f")

    def test_6424(self):
        "6424 - insert an int8 vector with 65,533 dimensions"
        value = array.array("b", [2] * 65533)
        self.__test_insert_and_fetch(value, "VectorFlexAllCol", "b")

    def test_6425(self):
        "6425 - insert vectors with different dimensions"
        for dim in [30, 70, 255, 256, 65534, 65535]:
            for typ in ["f", "d", "b"]:
                with self.subTest(dim=dim, typ=typ):
                    element_value = 3 if typ == "b" else 1.5
                    value = array.array(typ, [element_value] * dim)
                    self.__test_insert_and_fetch(
                        value, "VectorFlexAllCol", typ
                    )

    @unittest.skip("awaiting database support")
    def test_6426(self):
        "6426 - insert and fetch VECTOR data using CLOB"
        value = [6426, -15.75, 283.125, -8.625]
        clob = self.conn.createlob(oracledb.DB_TYPE_CLOB)
        clob.write(json.dumps(value))
        self.cursor.execute("delete from TestVectors")
        self.cursor.execute(
            """
            insert into TestVectors (IntCol, VectorFlexAllCol)
            values(1, :value)
            """,
            value=clob,
        )

        def type_handler(cursor, metadata):
            if metadata.name == "VECTORFLEXALLCOL":
                return cursor.var(
                    oracledb.DB_TYPE_CLOB, arraysize=cursor.arraysize
                )

        self.cursor.outputtypehandler = type_handler

        self.cursor.execute("select VectorFlexAllCol from TestVectors")
        (clob_data,) = self.cursor.fetchone()
        fetched_value = json.loads(clob_data.read())
        self.assertEqual(fetched_value, value)

    def test_6427(self):
        "6427 - insert and fetch VECTOR data using strings"
        value = [6427, -25.75, 383.125, -18.625]
        self.cursor.execute("delete from TestVectors")
        self.cursor.execute(
            """
            insert into TestVectors (IntCol, VectorFlexAllCol)
            values(1, :value)
            """,
            value=json.dumps(value),
        )

        def type_handler(cursor, metadata):
            if metadata.name == "VECTORFLEXALLCOL":
                return cursor.var(
                    oracledb.DB_TYPE_LONG, arraysize=cursor.arraysize
                )

        self.cursor.outputtypehandler = type_handler

        self.cursor.execute("select VectorFlexAllCol from TestVectors")
        (fetched_value,) = self.cursor.fetchone()
        self.assertEqual(json.loads(fetched_value), value)

    def test_6428(self):
        "6428 - insert vectors with flexible dimensions and conversion"
        for dim in [30, 255, 256, 257, 32768, 65535]:
            for source_type in ["f", "d", "b"]:
                for target_type in ["f", "d", "b"]:
                    with self.subTest(
                        dim=dim,
                        source_type=source_type,
                        target_type=target_type,
                    ):
                        if target_type == "f":
                            target_col = "VectorFlex32Col"
                        elif target_type == "d":
                            target_col = "VectorFlex64Col"
                        else:
                            target_col = "VectorFlex8Col"
                        element_value = 4 if source_type == "b" else 2.5
                        value = array.array(source_type, [element_value] * dim)
                        self.__test_insert_and_fetch(
                            value, target_col, target_type
                        )

    @unittest.skip("awaiting database support")
    def test_6429(self):
        "6427 - insert and fetch large VECTOR data using strings"
        value = [0.12345678925] * 35625
        self.cursor.execute("delete from TestVectors")
        self.cursor.execute(
            """
            insert into TestVectors (IntCol, VectorFlex64Col)
            values(1, :value)
            """,
            value=json.dumps(value),
        )

        def type_handler(cursor, metadata):
            if metadata.name == "VECTORFLEX64COL":
                return cursor.var(
                    oracledb.DB_TYPE_LONG, arraysize=cursor.arraysize
                )

        self.cursor.outputtypehandler = type_handler

        self.cursor.execute("select VectorFlex64Col from TestVectors")
        (fetched_value,) = self.cursor.fetchone()
        self.assertEqual(json.loads(fetched_value), value)

    def test_6430(self):
        "6430 - test binding a vector with inf values (negative)"
        value = array.array(
            "d", [float("inf") if i % 2 else float("-inf") for i in range(16)]
        )
        with self.assertRaisesFullCode("ORA-51805"):
            self.cursor.execute("select :1 from dual", [value])

    def test_6431(self):
        "6431 - test setting an invalid type to a vector"
        var = self.cursor.var(oracledb.DB_TYPE_VECTOR)
        self.assertRaises(TypeError, var.setvalue, 0, [[i] for i in range(16)])

    def test_6432(self):
        "6432 - fetch JSON value with an embedded vector"
        self.cursor.execute(
            """
            select json_object(
                'id' : 6432,
                'vector' : to_vector('[1, 2, 3]')
                returning json
            ) from dual
            """
        )
        (result,) = self.cursor.fetchone()
        expected_val = dict(id=6432, vector=array.array("f", [1, 2, 3]))
        self.assertEqual(result, expected_val)

    def test_6433(self):
        "6433 - bind JSON value with an embedded vector"
        value = dict(id=6433, vector=array.array("d", [6433, 6433.25, 6433.5]))
        self.cursor.execute("delete from TestJson")
        self.cursor.setinputsizes(None, oracledb.DB_TYPE_JSON)
        self.cursor.execute(
            "insert into TestJson values (:1, :2)", [6433, value]
        )
        self.conn.commit()
        self.cursor.execute("select JsonCol from TestJson")
        (fetched_val,) = self.cursor.fetchone()
        self.assertEqual(fetched_val, value)

    def test_6434(self):
        "6434 - executemany() without setinputsizes()"
        self.cursor.execute("delete from TestVectors")
        values = [array.array("f", [3.1416, 4]), [3.14159, 5]]
        self.cursor.executemany(
            """
            insert into TestVectors (IntCol, VectorFlexTypeCol)
            values (:1, :2)
            """,
            list(enumerate(values)),
        )
        self.cursor.execute(
            "select VectorFlexTypeCol from TestVectors order by IntCol"
        )
        expected_value = [
            (array.array("f", [3.1416, 4.0]),),
            (array.array("d", [3.14159, 5.0]),),
        ]
        self.assertEqual(self.cursor.fetchall(), expected_value)

    def test_6435(self):
        "6435 - executemany() with setinputsizes()"
        self.cursor.execute("delete from TestVectors")
        values = [[144, 1000], array.array("d", [66.0, 7.14])]
        self.cursor.setinputsizes(None, oracledb.DB_TYPE_VECTOR)
        self.cursor.executemany(
            """
            insert into TestVectors (IntCol, VectorFlex64Col)
            values (:1, :2)
            """,
            list(enumerate(values)),
        )
        self.cursor.execute(
            "select VectorFlex64Col from TestVectors order by IntCol"
        )
        expected_value = [
            (array.array("d", [144.0, 1000.0]),),
            (array.array("d", [66.0, 7.14]),),
        ]
        self.assertEqual(self.cursor.fetchall(), expected_value)

    def test_6436(self):
        "6436 - vector with zero dimensions"
        with self.assertRaisesFullCode("DPY-4031"):
            self.cursor.setinputsizes(oracledb.DB_TYPE_VECTOR)
            self.cursor.execute("select :1", [[]])
        with self.assertRaisesFullCode("DPY-4031"):
            self.cursor.execute("select :1", [array.array("d", [])])

    def test_6437(self):
        "6437 - insert a list vector into a flexible format column"
        value = [1.5, 9.9]
        self.__test_insert_and_fetch(value, "VectorFlexTypeCol", "d")

    def test_6438(self):
        "6438 - insert a list vector into a flexible size column"
        value = [1.5, 9.9, 3, 8, 4.25, 7, 5, 6.125, 0, 2, 6, 4, 5, 6, 7, 8]
        self.__test_insert_and_fetch(value, "VectorFlexAllCol", "d")

    def test_6439(self):
        "6439 - insert a list vector into a flexible float32 column"
        value = [1.5, 9.9, 3, 8, 4.25, 7, 5, 6.125, 0, 2, 6, 4, 5, 6, 7, 8]
        self.__test_insert_and_fetch(value, "VectorFlex32Col", "f")

    def test_6440(self):
        "6440 - insert a list vector into a flexible float64 column"
        value = [1.5, 9.9, 3, 8, 4.25, 7, 5, 6.125, 0, 2, 6, 4, 5, 6, 7, 8]
        self.__test_insert_and_fetch(value, "VectorFlex64Col", "d")

    def test_6441(self):
        "6441 - insert a list vector into a float32 column"
        value = [1.5, 9.9, 3, 8, 4.25, 7, 5, 6.125, 0, 2, 6, 4, 5, 6, 7, 8]
        self.__test_insert_and_fetch(value, "Vector32Col", "f")

    def test_6442(self):
        "6442 - insert a list vector into a float64 column"
        value = [1.5, 9.9, 3, 8, 4.25, 7, 5, 6.125, 0, 2, 6, 4, 5, 6, 7, 8]
        self.__test_insert_and_fetch(value, "Vector64Col", "d")

    def test_6443(self):
        "6443 - insert a list vector into a flexible int8 column"
        value = [1, 9, 3, 8, 4, 7, 5, 6, 0, 2, 6, 4, 5, 6, 7, 8]
        self.__test_insert_and_fetch(value, "VectorFlex8Col", "b")

    def test_6444(self):
        "6444 - insert a list vector into an int8 column"
        value = [1, 9, 3, 8, 4, 7, 5, 6, 0, 2, 6, 4, 5, 6, 7, 8]
        self.__test_insert_and_fetch(value, "Vector8Col", "b")

    def test_6445(self):
        "6445 - test setting a PL-SQL type to a float32 vector"
        vec1 = array.array("f", [1, 1.5, 2, 2.5])
        vec2 = array.array("f", [4, 4.5, 5, 5.5])
        self.__test_plsql_insert_and_fetch(vec1, vec2, 6)

        vec3 = array.array("f", [3.5] * 65535)
        vec4 = array.array("f", [2.5] * 65535)
        self.__test_plsql_insert_and_fetch(vec3, vec4, 256)

    def test_6446(self):
        "6446 - test setting a PL-SQL type to a float64 vector"
        vec1 = array.array("d", [1, 1.5, 2, 2.5])
        vec2 = array.array("d", [4, 4.5, 5, 5.5])
        self.__test_plsql_insert_and_fetch(vec1, vec2, 6)

        vec3 = array.array("d", [3.5] * 65535)
        vec4 = array.array("d", [2.5] * 65535)
        self.__test_plsql_insert_and_fetch(vec3, vec4, 256)

    def test_6447(self):
        "6447 - test setting a PL-SQL type to a int8 vector"
        vec1 = array.array("b", [1, 2, 3, 4])
        vec2 = array.array("b", [5, 6, 7, 8])
        self.__test_plsql_insert_and_fetch(vec1, vec2, 8)

        vec3 = array.array("b", [3] * 65535)
        vec4 = array.array("b", [2] * 65535)
        self.__test_plsql_insert_and_fetch(vec3, vec4, 256)


if __name__ == "__main__":
    test_env.run_test_cases()



#------------------------------------------------------------------------------
# Copyright (c) 2025, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# aq_deq.pyx
#
# Cython file defining the messages that are sent to the database and the
# responses that are received by the client for dequeuing an AQ message
# (embedded in thin_impl.pyx).
#------------------------------------------------------------------------------

@cython.final
cdef class AqDeqMessage(AqBaseMessage):
    cdef:
        ThinMsgPropsImpl props_impl

    cdef int _initialize_hook(self) except -1:
        """
        Perform initialization
        """
        self.function_code = TNS_FUNC_AQ_DEQ

    cdef int _process_return_parameters(self, ReadBuffer buf) except -1:
        """
        Process the return parameters of the AQ Dequeue request.
        """
        cdef:
            uint32_t num_bytes
            uint16_t keyword
            uint32_t imageLength
            ThinDbObjectImpl obj_impl
            ThinDbObjectTypeImpl type_impl
        buf.read_ub4(&num_bytes)
        if num_bytes > 0:
            self._process_msg_props(buf, self.props_impl)
            self.props_impl.recipients = self._process_recipients(buf)
            self.props_impl.payload = self._process_payload(buf)
            self.props_impl.msgid = self._process_msg_id(buf)

    cdef int _write_message(self, WriteBuffer buf) except -1:
        """
        Write message to the network buffers.
        """
        cdef:
             bytes queue_name_bytes
             bytes consumer_name_bytes
             bytes correlation_bytes
             bytes condition_bytes
             uint16_t delivery_mode
             int deq_flags
        self._write_function_code(buf)
        queue_name_bytes = self.queue_impl.name.encode()
        buf.write_uint8(1)                      # queue name (pointer)
        buf.write_ub4(len(queue_name_bytes))    # queue name length
        buf.write_uint8(1)                      # message properties
        buf.write_uint8(1)                      # msg props length
        buf.write_uint8(1)                      # recipient list
        buf.write_uint8(1)                      # recipient list length
        if self.deq_options_impl.consumer_name:
            consumer_name_bytes = self.deq_options_impl.consumer_name.encode()
            buf.write_uint8(1)                  # consumer name
            buf.write_ub4(len(consumer_name_bytes))
        else:
            consumer_name_bytes = None
            buf.write_uint8(0)                  # consumer name
            buf.write_ub4(0)                    # consumer name length
        buf.write_sb4(self.deq_options_impl.mode) # dequeue mode
        buf.write_sb4(self.deq_options_impl.navigation) # navigation
        buf.write_sb4(self.deq_options_impl.visibility) # visibility
        buf.write_sb4(self.deq_options_impl.wait) # wait
        if self.deq_options_impl.msgid:
            buf.write_uint8(1)                  # select mesg id
            buf.write_ub4(TNS_AQ_MESSAGE_ID_LENGTH) # mesg id len
        else:
            buf.write_uint8(0)                  # select mesg id
            buf.write_ub4(0)                    # select mesg id length
        if self.deq_options_impl.correlation:
            correlation_bytes = self.deq_options_impl.correlation.encode()
            buf.write_uint8(1)                  # correlation id
            buf.write_ub4(len(correlation_bytes)) # correlation id len
        else:
            correlation_bytes = None
            buf.write_uint8(0)                  # correlation id
            buf.write_ub4(0)                    # correlation id len
        buf.write_uint8(1)                      # toid of payload
        buf.write_ub4(16)                       # toid length
        buf.write_ub2(TNS_AQ_MESSAGE_VERSION)
        buf.write_uint8(1)                      # payload
        buf.write_uint8(1)                      # return msg id
        buf.write_ub4(TNS_AQ_MESSAGE_ID_LENGTH)
        deq_flags = 0
        delivery_mode = self.deq_options_impl.delivery_mode
        if (delivery_mode == TNS_AQ_MSG_BUFFERED):
            deq_flags |= TNS_KPD_AQ_BUFMSG
        elif (delivery_mode == TNS_AQ_MSG_PERSISTENT_OR_BUFFERED):
            deq_flags |= TNS_KPD_AQ_EITHER
        buf.write_ub4(deq_flags)                # dequeue flags
        if self.deq_options_impl.condition:
            condition_bytes = self.deq_options_impl.condition.encode()
            buf.write_uint8(1)                  # condition (pointer)
            buf.write_ub4(len(condition_bytes)) # condition length
        else:
            condition_bytes = None
            buf.write_uint8(0)                  # condition
            buf.write_ub4(0)                    # condition length
        buf.write_uint8(0)                      # extensions
        buf.write_ub4(0)                        # number of extensions
        if buf._caps.ttc_field_version >= TNS_CCAP_FIELD_VERSION_20_1:
            buf.write_uint8(0)                  # JSON payload
        if buf._caps.ttc_field_version >= TNS_CCAP_FIELD_VERSION_21_1:
            buf.write_ub4(-1)                   # shard id

        buf.write_bytes_with_length(queue_name_bytes)
        if consumer_name_bytes is not None:
            buf.write_bytes_with_length(consumer_name_bytes)
        if self.deq_options_impl.msgid:
            buf.write_bytes(self.deq_options_impl.msgid)
        if correlation_bytes is not None:
            buf.write_bytes_with_length(correlation_bytes)
        buf.write_bytes(self.queue_impl.payload_toid)
        if condition_bytes is not None:
            buf.write_bytes_with_length(condition_bytes)



#------------------------------------------------------------------------------
# Copyright (c) 2020, 2025, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# constants.pxi
#
# Cython file defining constants used by the thin implementation classes
# (embedded in thin_impl.pyx).
#------------------------------------------------------------------------------

# packet types
cdef enum:
    TNS_PACKET_TYPE_CONNECT = 1
    TNS_PACKET_TYPE_ACCEPT = 2
    TNS_PACKET_TYPE_REFUSE = 4
    TNS_PACKET_TYPE_DATA = 6
    TNS_PACKET_TYPE_RESEND = 11
    TNS_PACKET_TYPE_MARKER = 12
    TNS_PACKET_TYPE_CONTROL = 14
    TNS_PACKET_TYPE_REDIRECT = 5

# packet flags
cdef enum:
    TNS_PACKET_FLAG_REDIRECT = 0x04
    TNS_PACKET_FLAG_TLS_RENEG = 0x08

# data flags
cdef enum:
    TNS_DATA_FLAGS_BEGIN_PIPELINE = 0x1000
    TNS_DATA_FLAGS_END_OF_REQUEST = 0x800
    TNS_DATA_FLAGS_END_OF_RESPONSE = 0x2000
    TNS_DATA_FLAGS_EOF = 0x0040

# marker types
cdef enum:
    TNS_MARKER_TYPE_BREAK = 1
    TNS_MARKER_TYPE_RESET = 2
    TNS_MARKER_TYPE_INTERRUPT = 3

# AQ delivery modes
cdef enum:
    TNS_AQ_MSG_BUFFERED = 2
    TNS_AQ_MSG_PERSISTENT = 1
    TNS_AQ_MSG_PERSISTENT_OR_BUFFERED = 3

# AQ dequeue modes
cdef enum:
    TNS_AQ_DEQ_BROWSE = 1
    TNS_AQ_DEQ_LOCKED = 2
    TNS_AQ_DEQ_REMOVE = 3
    TNS_AQ_DEQ_REMOVE_NODATA = 4

# AQ dequeue navigation modes
cdef enum:
    TNS_AQ_DEQ_FIRST_MSG = 1
    TNS_AQ_DEQ_NEXT_MSG = 3
    TNS_AQ_DEQ_NEXT_TRANSACTION = 2

# AQ dequeue visibility modes
cdef enum:
    TNS_AQ_DEQ_IMMEDIATE = 1
    TNS_AQ_DEQ_ON_COMMIT = 2

# AQ dequeue wait modes
cdef enum:
    TNS_AQ_DEQ_NO_WAIT = 0
    TNS_AQ_DEQ_WAIT_FOREVER = <uint32_t>-1

# AQ enqueue visibility modes
cdef enum:
    TNS_AQ_ENQ_IMMEDIATE = 1
    TNS_AQ_ENQ_ON_COMMIT = 2

# AQ message states
cdef enum:
    TNS_AQ_MSG_EXPIRED = 3
    TNS_AQ_MSG_PROCESSED = 2
    TNS_AQ_MSG_READY = 0
    TNS_AQ_MSG_WAITING = 1

# AQ other constants
cdef enum:
    TNS_AQ_MSG_NO_DELAY = 0
    TNS_AQ_MSG_NO_EXPIRATION = -1
    TNS_AQ_ARRAY_ENQ = 0x01
    TNS_AQ_ARRAY_DEQ = 0x02
    TNS_AQ_ARRAY_FLAGS_RETURN_MESSAGE_ID = 0x01
    TNS_TTC_ENQ_STREAMING_ENABLED = 0x00000001
    TNS_TTC_ENQ_STREAMING_DISABLED = 0x00000000

# AQ flags
cdef enum:
    TNS_KPD_AQ_BUFMSG = 0x02
    TNS_KPD_AQ_EITHER = 0x10

# errors
cdef enum:
    TNS_ERR_INCONSISTENT_DATA_TYPES = 932
    TNS_ERR_VAR_NOT_IN_SELECT_LIST = 1007
    TNS_ERR_INBAND_MESSAGE = 12573
    TNS_ERR_INVALID_SERVICE_NAME = 12514
    TNS_ERR_INVALID_SID = 12505
    TNS_ERR_NO_DATA_FOUND = 1403
    TNS_ERR_SESSION_SHUTDOWN = 12572
    TNS_ERR_ARRAY_DML_ERRORS = 24381
    TNS_ERR_EXCEEDED_IDLE_TIME = 2396
    TNS_ERR_NO_MESSAGES_FOUND = 25228

# message types
cdef enum:
    TNS_MSG_TYPE_PROTOCOL = 1
    TNS_MSG_TYPE_DATA_TYPES = 2
    TNS_MSG_TYPE_FUNCTION = 3
    TNS_MSG_TYPE_ERROR = 4
    TNS_MSG_TYPE_ROW_HEADER = 6
    TNS_MSG_TYPE_ROW_DATA = 7
    TNS_MSG_TYPE_PARAMETER = 8
    TNS_MSG_TYPE_STATUS = 9
    TNS_MSG_TYPE_IO_VECTOR = 11
    TNS_MSG_TYPE_LOB_DATA = 14
    TNS_MSG_TYPE_WARNING = 15
    TNS_MSG_TYPE_DESCRIBE_INFO = 16
    TNS_MSG_TYPE_PIGGYBACK = 17
    TNS_MSG_TYPE_FLUSH_OUT_BINDS = 19
    TNS_MSG_TYPE_BIT_VECTOR = 21
    TNS_MSG_TYPE_SERVER_SIDE_PIGGYBACK = 23
    TNS_MSG_TYPE_ONEWAY_FN = 26
    TNS_MSG_TYPE_IMPLICIT_RESULTSET = 27
    TNS_MSG_TYPE_RENEGOTIATE = 28
    TNS_MSG_TYPE_END_OF_RESPONSE = 29
    TNS_MSG_TYPE_TOKEN = 33
    TNS_MSG_TYPE_FAST_AUTH = 34

# parameter keyword numbers
cdef enum:
    TNS_KEYWORD_NUM_CURRENT_SCHEMA = 168
    TNS_KEYWORD_NUM_EDITION = 172

# bind flags
cdef enum:
    TNS_BIND_USE_INDICATORS = 0x0001
    TNS_BIND_ARRAY = 0x0040

# bind directions
cdef enum:
    TNS_BIND_DIR_OUTPUT = 16
    TNS_BIND_DIR_INPUT = 32
    TNS_BIND_DIR_INPUT_OUTPUT = 48

# database object image flags
cdef enum:
    TNS_OBJ_IS_VERSION_81 = 0x80
    TNS_OBJ_IS_DEGENERATE = 0x10
    TNS_OBJ_IS_COLLECTION = 0x08
    TNS_OBJ_NO_PREFIX_SEG = 0x04
    TNS_OBJ_IMAGE_VERSION = 1

# database object flags
cdef enum:
    TNS_OBJ_MAX_SHORT_LENGTH = 245
    TNS_OBJ_ATOMIC_NULL = 253
    TNS_OBJ_NON_NULL_OID = 0x02
    TNS_OBJ_HAS_EXTENT_OID = 0x08
    TNS_OBJ_TOP_LEVEL = 0x01
    TNS_OBJ_HAS_INDEXES = 0x10

# database object collection types
cdef enum:
    TNS_OBJ_PLSQL_INDEX_TABLE = 1
    TNS_OBJ_NESTED_TABLE = 2
    TNS_OBJ_VARRAY = 3

# database object TDS type codes
cdef enum:
    TNS_OBJ_TDS_TYPE_CHAR = 1
    TNS_OBJ_TDS_TYPE_DATE = 2
    TNS_OBJ_TDS_TYPE_FLOAT = 5
    TNS_OBJ_TDS_TYPE_NUMBER = 6
    TNS_OBJ_TDS_TYPE_VARCHAR = 7
    TNS_OBJ_TDS_TYPE_BOOLEAN = 8
    TNS_OBJ_TDS_TYPE_RAW = 19
    TNS_OBJ_TDS_TYPE_TIMESTAMP = 21
    TNS_OBJ_TDS_TYPE_TIMESTAMP_TZ = 23
    TNS_OBJ_TDS_TYPE_OBJ = 27
    TNS_OBJ_TDS_TYPE_COLL = 28
    TNS_OBJ_TDS_TYPE_CLOB = 29
    TNS_OBJ_TDS_TYPE_BLOB = 30
    TNS_OBJ_TDS_TYPE_TIMESTAMP_LTZ = 33
    TNS_OBJ_TDS_TYPE_BINARY_FLOAT = 37
    TNS_OBJ_TDS_TYPE_START_EMBED_ADT = 39
    TNS_OBJ_TDS_TYPE_END_EMBED_ADT = 40
    TNS_OBJ_TDS_TYPE_SUBTYPE_MARKER = 43
    TNS_OBJ_TDS_TYPE_EMBED_ADT_INFO = 44
    TNS_OBJ_TDS_TYPE_BINARY_DOUBLE = 45

# xml type constants
cdef enum:
    TNS_XML_TYPE_LOB = 0x0001
    TNS_XML_TYPE_STRING = 0x0004
    TNS_XML_TYPE_FLAG_SKIP_NEXT_4 = 0x100000

# execute options
cdef enum:
    TNS_EXEC_OPTION_PARSE = 0x01
    TNS_EXEC_OPTION_BIND = 0x08
    TNS_EXEC_OPTION_DEFINE = 0x10
    TNS_EXEC_OPTION_EXECUTE = 0x20
    TNS_EXEC_OPTION_FETCH = 0x40
    TNS_EXEC_OPTION_COMMIT = 0x100
    TNS_EXEC_OPTION_COMMIT_REEXECUTE = 0x1
    TNS_EXEC_OPTION_PLSQL_BIND = 0x400
    TNS_EXEC_OPTION_NOT_PLSQL = 0x8000
    TNS_EXEC_OPTION_DESCRIBE = 0x20000
    TNS_EXEC_OPTION_NO_COMPRESSED_FETCH = 0x40000
    TNS_EXEC_OPTION_BATCH_ERRORS = 0x80000

# execute flags
cdef enum:
    TNS_EXEC_FLAGS_DML_ROWCOUNTS = 0x4000
    TNS_EXEC_FLAGS_IMPLICIT_RESULTSET = 0x8000
    TNS_EXEC_FLAGS_SCROLLABLE = 0x02

# fetch orientations
cdef enum:
    TNS_FETCH_ORIENTATION_ABSOLUTE = 0x20
    TNS_FETCH_ORIENTATION_CURRENT = 0x01
    TNS_FETCH_ORIENTATION_FIRST = 0x04
    TNS_FETCH_ORIENTATION_LAST = 0x08
    TNS_FETCH_ORIENTATION_NEXT = 0x02
    TNS_FETCH_ORIENTATION_PRIOR = 0x10
    TNS_FETCH_ORIENTATION_RELATIVE = 0x40

# server side piggyback op codes
cdef enum:
    TNS_SERVER_PIGGYBACK_QUERY_CACHE_INVALIDATION = 1
    TNS_SERVER_PIGGYBACK_OS_PID_MTS = 2
    TNS_SERVER_PIGGYBACK_TRACE_EVENT = 3
    TNS_SERVER_PIGGYBACK_SESS_RET = 4
    TNS_SERVER_PIGGYBACK_SYNC = 5
    TNS_SERVER_PIGGYBACK_LTXID = 7
    TNS_SERVER_PIGGYBACK_AC_REPLAY_CONTEXT = 8
    TNS_SERVER_PIGGYBACK_EXT_SYNC = 9
    TNS_SERVER_PIGGYBACK_SESS_SIGNATURE = 10

# session return constants
cdef enum:
    TNS_SESSGET_SESSION_CHANGED = 4

# LOB operations
cdef enum:
    TNS_LOB_OP_GET_LENGTH = 0x0001
    TNS_LOB_OP_READ = 0x0002
    TNS_LOB_OP_TRIM = 0x0020
    TNS_LOB_OP_WRITE = 0x0040
    TNS_LOB_OP_GET_CHUNK_SIZE = 0x4000
    TNS_LOB_OP_CREATE_TEMP = 0x0110
    TNS_LOB_OP_FREE_TEMP = 0x0111
    TNS_LOB_OP_OPEN = 0x8000
    TNS_LOB_OP_CLOSE = 0x10000
    TNS_LOB_OP_IS_OPEN = 0x11000
    TNS_LOB_OP_ARRAY = 0x80000
    TNS_LOB_OP_FILE_EXISTS = 0x0800
    TNS_LOB_OP_FILE_OPEN = 0x0100
    TNS_LOB_OP_FILE_CLOSE = 0x0200
    TNS_LOB_OP_FILE_ISOPEN = 0x0400

# LOB locator constants
cdef enum:
    TNS_LOB_LOC_OFFSET_FLAG_1 = 4
    TNS_LOB_LOC_OFFSET_FLAG_3 = 6
    TNS_LOB_LOC_OFFSET_FLAG_4 = 7
    TNS_LOB_QLOCATOR_VERSION = 4
    TNS_LOB_LOC_FIXED_OFFSET = 16

# LOB locator flags (byte 1)
cdef enum:
    TNS_LOB_LOC_FLAGS_BLOB = 0x01
    TNS_LOB_LOC_FLAGS_VALUE_BASED = 0x20
    TNS_LOB_LOC_FLAGS_ABSTRACT = 0x40

# LOB locator flags (byte 2)
cdef enum:
    TNS_LOB_LOC_FLAGS_INIT = 0x08

# LOB locator flags (byte 4)
cdef enum:
    TNS_LOB_LOC_FLAGS_TEMP = 0x01
    TNS_LOB_LOC_FLAGS_VAR_LENGTH_CHARSET = 0x80

# other LOB constants
cdef enum:
    TNS_LOB_OPEN_READ_WRITE = 2
    TNS_LOB_OPEN_READ_ONLY = 11
    TNS_LOB_PREFETCH_FLAG = 0x2000000

# end-to-end metrics
cdef enum:
    TNS_END_TO_END_ACTION = 0x0010
    TNS_END_TO_END_CLIENT_IDENTIFIER = 0x0001
    TNS_END_TO_END_CLIENT_INFO = 0x0100
    TNS_END_TO_END_DBOP = 0x0200
    TNS_END_TO_END_MODULE = 0x0008

# versions
cdef enum:
    TNS_VERSION_DESIRED = 319
    TNS_VERSION_MINIMUM = 300
    TNS_VERSION_MIN_ACCEPTED = 315      # 12.1
    TNS_VERSION_MIN_LARGE_SDU = 315
    TNS_VERSION_MIN_OOB_CHECK = 318
    TNS_VERSION_MIN_END_OF_RESPONSE = 319

# control packet types
cdef enum:
    TNS_CONTROL_TYPE_INBAND_NOTIFICATION = 8
    TNS_CONTROL_TYPE_RESET_OOB = 9

# connect flags
cdef enum:
    TNS_GSO_DONT_CARE = 0x0001
    TNS_GSO_CAN_RECV_ATTENTION = 0x0400
    TNS_NSI_NA_REQUIRED = 0x10
    TNS_NSI_DISABLE_NA = 0x04
    TNS_NSI_SUPPORT_SECURITY_RENEG = 0x80

# other connection constants
cdef enum:
    TNS_PROTOCOL_CHARACTERISTICS = 0x4f98
    TNS_CHECK_OOB = 0x01

# TTC functions
cdef enum:
    TNS_FUNC_AUTH_PHASE_ONE = 118
    TNS_FUNC_AUTH_PHASE_TWO = 115
    TNS_FUNC_CLOSE_CURSORS = 105
    TNS_FUNC_COMMIT = 14
    TNS_FUNC_EXECUTE = 94
    TNS_FUNC_FETCH = 5
    TNS_FUNC_LOB_OP = 96
    TNS_FUNC_AQ_ENQ = 121
    TNS_FUNC_AQ_DEQ = 122
    TNS_FUNC_ARRAY_AQ = 145
    TNS_FUNC_LOGOFF = 9
    TNS_FUNC_PING = 147
    TNS_FUNC_PIPELINE_BEGIN = 199
    TNS_FUNC_PIPELINE_END = 200
    TNS_FUNC_ROLLBACK = 15
    TNS_FUNC_SET_END_TO_END_ATTR = 135
    TNS_FUNC_REEXECUTE = 4
    TNS_FUNC_REEXECUTE_AND_FETCH = 78
    TNS_FUNC_SESSION_GET = 162
    TNS_FUNC_SESSION_RELEASE = 163
    TNS_FUNC_SESSION_STATE = 176
    TNS_FUNC_SET_SCHEMA = 152
    TNS_FUNC_TPC_TXN_SWITCH = 103
    TNS_FUNC_TPC_TXN_CHANGE_STATE = 104

# TTC authentication modes
cdef enum:
    TNS_AUTH_MODE_LOGON = 0x00000001
    TNS_AUTH_MODE_CHANGE_PASSWORD = 0x00000002
    TNS_AUTH_MODE_SYSDBA = 0x00000020
    TNS_AUTH_MODE_SYSOPER = 0x00000040
    TNS_AUTH_MODE_WITH_PASSWORD = 0x00000100
    TNS_AUTH_MODE_SYSASM = 0x00400000
    TNS_AUTH_MODE_SYSBKP = 0x01000000
    TNS_AUTH_MODE_SYSDGD = 0x02000000
    TNS_AUTH_MODE_SYSKMT = 0x04000000
    TNS_AUTH_MODE_SYSRAC = 0x08000000
    TNS_AUTH_MODE_IAM_TOKEN = 0x20000000

# character sets and encodings
cdef enum:
    TNS_CHARSET_UTF8 = 873
    TNS_CHARSET_UTF16 = 2000
    TNS_ENCODING_MULTI_BYTE = 0x01
    TNS_ENCODING_CONV_LENGTH = 0x02

# compile time capability indices
cdef enum:
    TNS_CCAP_SQL_VERSION = 0
    TNS_CCAP_LOGON_TYPES = 4
    TNS_CCAP_FEATURE_BACKPORT = 5
    TNS_CCAP_FIELD_VERSION = 7
    TNS_CCAP_SERVER_DEFINE_CONV = 8
    TNS_CCAP_DEQUEUE_WITH_SELECTOR = 9
    TNS_CCAP_TTC1 = 15
    TNS_CCAP_OCI1 = 16
    TNS_CCAP_TDS_VERSION = 17
    TNS_CCAP_RPC_VERSION = 18
    TNS_CCAP_RPC_SIG = 19
    TNS_CCAP_DBF_VERSION = 21
    TNS_CCAP_LOB = 23
    TNS_CCAP_TTC2 = 26
    TNS_CCAP_UB2_DTY = 27
    TNS_CCAP_OCI2 = 31
    TNS_CCAP_CLIENT_FN = 34
    TNS_CCAP_TTC3 = 37
    TNS_CCAP_SESS_SIGNATURE_VERSION = 39
    TNS_CCAP_TTC4 = 40
    TNS_CCAP_LOB2 = 42
    TNS_CCAP_TTC5 = 44
    TNS_CCAP_VECTOR_FEATURES = 52
    TNS_CCAP_MAX = 53

# compile time capability values
cdef enum:
    TNS_CCAP_SQL_VERSION_MAX = 6
    TNS_CCAP_FIELD_VERSION_11_2 = 6
    TNS_CCAP_FIELD_VERSION_12_1 = 7
    TNS_CCAP_FIELD_VERSION_12_2 = 8
    TNS_CCAP_FIELD_VERSION_12_2_EXT1 = 9
    TNS_CCAP_FIELD_VERSION_18_1 = 10
    TNS_CCAP_FIELD_VERSION_18_1_EXT_1 = 11
    TNS_CCAP_FIELD_VERSION_19_1 = 12
    TNS_CCAP_FIELD_VERSION_19_1_EXT_1 = 13
    TNS_CCAP_FIELD_VERSION_20_1 = 14
    TNS_CCAP_FIELD_VERSION_20_1_EXT_1 = 15
    TNS_CCAP_FIELD_VERSION_21_1 = 16
    TNS_CCAP_FIELD_VERSION_23_1 = 17
    TNS_CCAP_FIELD_VERSION_23_1_EXT_1 = 18
    TNS_CCAP_FIELD_VERSION_23_1_EXT_2 = 19
    TNS_CCAP_FIELD_VERSION_23_1_EXT_3 = 20
    TNS_CCAP_FIELD_VERSION_23_1_EXT_4 = 21
    TNS_CCAP_FIELD_VERSION_23_1_EXT_5 = 22
    TNS_CCAP_FIELD_VERSION_23_3_EXT_6 = 23
    TNS_CCAP_FIELD_VERSION_23_4 = 24
    TNS_CCAP_FIELD_VERSION_MAX = 24
    TNS_CCAP_O5LOGON = 8
    TNS_CCAP_O5LOGON_NP = 2
    TNS_CCAP_O7LOGON = 32
    TNS_CCAP_O8LOGON_LONG_IDENTIFIER = 64
    TNS_CCAP_O9LOGON_LONG_PASSWORD = 0x80
    TNS_CCAP_CTB_IMPLICIT_POOL = 0x08
    TNS_CCAP_END_OF_CALL_STATUS = 0x01
    TNS_CCAP_IND_RCD = 0x08
    TNS_CCAP_FAST_BVEC = 0x20
    TNS_CCAP_FAST_SESSION_PROPAGATE = 0x10
    TNS_CCAP_APP_CTX_PIGGYBACK = 0x80
    TNS_CCAP_TDS_VERSION_MAX = 3
    TNS_CCAP_RPC_VERSION_MAX = 7
    TNS_CCAP_RPC_SIG_VALUE = 3
    TNS_CCAP_DBF_VERSION_MAX = 1
    TNS_CCAP_LTXID = 0x08
    TNS_CCAP_IMPLICIT_RESULTS = 0x10
    TNS_CCAP_BIG_CHUNK_CLR = 0x20
    TNS_CCAP_KEEP_OUT_ORDER = 0x80
    TNS_CCAP_LOB_UB8_SIZE = 0x01
    TNS_CCAP_LOB_ENCS = 0x02
    TNS_CCAP_LOB_PREFETCH_DATA = 0x04
    TNS_CCAP_LOB_TEMP_SIZE = 0x08
    TNS_CCAP_LOB_PREFETCH_LENGTH = 0x40
    TNS_CCAP_LOB_12C = 0x80
    TNS_CCAP_LOB2_QUASI = 0x01
    TNS_CCAP_LOB2_2GB_PREFETCH = 0x04
    TNS_CCAP_DRCP = 0x10
    TNS_CCAP_ZLNP = 0x04
    TNS_CCAP_INBAND_NOTIFICATION = 0x04
    TNS_CCAP_EXPLICIT_BOUNDARY = 0x40
    TNS_CCAP_END_OF_RESPONSE = 0x20
    TNS_CCAP_CLIENT_FN_MAX = 12
    TNS_CCAP_VECTOR_SUPPORT = 0x08
    TNS_CCAP_TOKEN_SUPPORTED = 0x02
    TNS_CCAP_PIPELINING_SUPPORT = 0x04
    TNS_CCAP_PIPELINING_BREAK = 0x10
    TNS_CCAP_VECTOR_FEATURE_BINARY = 0x01
    TNS_CCAP_VECTOR_FEATURE_SPARSE = 0x02

# runtime capability indices
cdef enum:
    TNS_RCAP_COMPAT = 0
    TNS_RCAP_TTC = 6
    TNS_RCAP_MAX = 11

# runtime capability values
cdef enum:
    TNS_RCAP_COMPAT_81 = 2
    TNS_RCAP_TTC_ZERO_COPY = 0x01
    TNS_RCAP_TTC_32K = 0x04
    TNS_RCAP_TTC_SESSION_STATE_OPS = 0x10

# verifier types
cdef enum:
    TNS_VERIFIER_TYPE_11G_1 = 0xb152
    TNS_VERIFIER_TYPE_11G_2 = 0x1b25
    TNS_VERIFIER_TYPE_12C = 0x4815

# UDS flags
cdef enum:
    TNS_UDS_FLAGS_IS_JSON = 0x00000100
    TNS_UDS_FLAGS_IS_OSON = 0x00000800

# end of call status flags
cdef enum:
    TNS_EOCS_FLAGS_TXN_IN_PROGRESS = 0x00000002
    TNS_EOCS_FLAGS_SESS_RELEASE = 0x00008000

# accept flags
cdef enum:
    TNS_ACCEPT_FLAG_CHECK_OOB = 0x00000001
    TNS_ACCEPT_FLAG_FAST_AUTH = 0x10000000
    TNS_ACCEPT_FLAG_HAS_END_OF_RESPONSE = 0x02000000

# transaction switching op codes
cdef enum:
    TNS_TPC_TXN_START = 0x01
    TNS_TPC_TXN_DETACH = 0x02

# transaction change state op codes
cdef enum:
    TNS_TPC_TXN_COMMIT = 0x01
    TNS_TPC_TXN_ABORT = 0x02
    TNS_TPC_TXN_PREPARE = 0x03
    TNS_TPC_TXN_FORGET = 0x04

# transaction states
cdef enum:
    TNS_TPC_TXN_STATE_PREPARE = 0
    TNS_TPC_TXN_STATE_REQUIRES_COMMIT = 1
    TNS_TPC_TXN_STATE_COMMITTED = 2
    TNS_TPC_TXN_STATE_ABORTED = 3
    TNS_TPC_TXN_STATE_READ_ONLY = 4
    TNS_TPC_TXN_STATE_FORGOTTEN = 5

# pipeline modes
cdef enum:
    TNS_PIPELINE_MODE_CONTINUE_ON_ERROR = 1
    TNS_PIPELINE_MODE_ABORT_ON_ERROR = 2

# AQ extension keywords
cdef enum:
    TNS_AQ_EXT_KEYWORD_AGENT_NAME = 64
    TNS_AQ_EXT_KEYWORD_AGENT_ADDRESS = 65
    TNS_AQ_EXT_KEYWORD_AGENT_PROTOCOL = 66
    TNS_AQ_EXT_KEYWORD_ORIGINAL_MSGID = 69

# session state flags
cdef enum:
    TNS_SESSION_STATE_REQUEST_BEGIN = 0x04
    TNS_SESSION_STATE_REQUEST_END = 0x08
    TNS_SESSION_STATE_EXPLICIT_BOUNDARY = 0x40

# other constants
cdef enum:
    TNS_ESCAPE_CHAR = 253
    TNS_MAX_ROWID_LENGTH = 18
    TNS_DURATION_SESSION = 10
    TNS_MAX_LONG_LENGTH = 0x7fffffff
    TNS_MAX_CONNECT_DATA = 230
    TNS_MAX_UROWID_LENGTH = 5267
    TNS_SERVER_CONVERTS_CHARS = 0x01
    TNS_JSON_MAX_LENGTH = 32 * 1024 * 1024
    TNS_VECTOR_MAX_LENGTH = 1 * 1024 * 1024
    TNS_AQ_MESSAGE_ID_LENGTH = 16
    TNS_AQ_MESSAGE_VERSION = 1

# base 64 encoding alphabet
cdef bytes TNS_BASE64_ALPHABET = \
        b'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/'
cdef bytearray TNS_BASE64_ALPHABET_ARRAY = \
        bytearray(TNS_BASE64_ALPHABET)
cdef bytes TNS_EXTENT_OID = bytes.fromhex('00000000000000000000000000010001')

# drcp release mode
cdef enum:
    DRCP_DEAUTHENTICATE = 0x00000002



# -----------------------------------------------------------------------------
# Copyright (c) 2024, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
E2100 - Module for testing BFILE. No special setup is required but the tests
here will only be run if the local_database is enabled.
"""

import os
import tempfile
import unittest

import oracledb
import test_env


@unittest.skipUnless(
    test_env.get_extended_config_bool("local_database"),
    "extended configuration local_database is disabled",
)
class TestCase(test_env.BaseTestCase):
    dir_name = "EXT_TEST_2100_DIR"

    def _setup_directory(self, local_name):
        """
        Setups the directory using the given local name.
        """
        user = test_env.get_main_user()
        with test_env.get_admin_connection() as conn:
            with conn.cursor() as cursor:
                cursor.execute(
                    f"""
                    create or replace directory {self.dir_name}
                    as '{local_name}'
                    """
                )
                cursor.execute(
                    f"grant read on directory {self.dir_name} to {user}"
                )

    def test_ext_2100(self):
        "E2100 - test fileexists() and getfilename()"
        with tempfile.TemporaryDirectory() as temp_dir:
            self._setup_directory(temp_dir)
            file_name = "test_2100.txt"
            contents = b"Some arbitrary data for test 2100"
            with open(os.path.join(temp_dir, file_name), "wb") as f:
                f.write(contents)
            self.cursor.execute(
                "select bfilename(:1, :2) from dual",
                [self.dir_name, file_name],
            )
            (bfile,) = self.cursor.fetchone()
            self.assertEqual(bfile.read(), contents)
            self.assertTrue(bfile.fileexists())
            self.assertEqual(bfile.getfilename(), (self.dir_name, file_name))

    def test_ext_2101(self):
        "E2101 - test setfilename()"
        with tempfile.TemporaryDirectory() as temp_dir:
            self._setup_directory(temp_dir)
            file_name = "test_2101.txt"
            contents = b"Some arbitrary data for test 2101"
            new_file_name = "test_2101b.txt"
            new_contents = b"Some arbitrary different data for test 2101"
            with open(os.path.join(temp_dir, file_name), "wb") as f:
                f.write(contents)
            with open(os.path.join(temp_dir, new_file_name), "wb") as f:
                f.write(new_contents)
            self.cursor.execute(
                "select bfilename(:1, :2) from dual",
                [self.dir_name, file_name],
            )
            (bfile,) = self.cursor.fetchone()
            self.assertEqual(bfile.read(), contents)
            bfile.setfilename(self.dir_name, new_file_name)
            self.assertEqual(bfile.read(), new_contents)

    def test_ext_2102(self):
        "E2102 - test BFILE with LOB methods"
        with tempfile.TemporaryDirectory() as temp_dir:
            self._setup_directory(temp_dir)
            file_name = "test_2102.txt"
            contents = b"Some arbitrary data for test 2102"
            with open(os.path.join(temp_dir, file_name), "wb") as f:
                f.write(contents)
            self.cursor.execute(
                "select bfilename(:1, :2) from dual",
                [self.dir_name, file_name],
            )
            (bfile,) = self.cursor.fetchone()
            self.assertEqual(bfile.size(), len(contents))
            self.assertEqual(bfile.read(7), contents[6:])
            self.assertEqual(bfile.read(7, 2), contents[6:8])
            self.assertFalse(bfile.isopen())
            bfile.open()
            self.assertTrue(bfile.isopen())
            bfile.close()
            self.assertFalse(bfile.isopen())
            with self.assertRaisesFullCode("DPY-3025"):
                bfile.getchunksize()
            with self.assertRaisesFullCode("DPY-3025"):
                bfile.trim(1)
            with self.assertRaisesFullCode("DPY-3025"):
                bfile.write("1")

    def test_ext_2103(self):
        "E2103 - test binding a BFILE"
        with tempfile.TemporaryDirectory() as temp_dir:
            self._setup_directory(temp_dir)
            file_name = "test_2103.txt"
            contents = b"Some arbitrary data for test 2103"
            with open(os.path.join(temp_dir, file_name), "wb") as f:
                f.write(contents)
            self.cursor.execute(
                "select bfilename(:1, :2) from dual",
                [self.dir_name, file_name],
            )
            (bfile,) = self.cursor.fetchone()
            self.cursor.execute("truncate table TestBfiles")
            self.cursor.execute(
                "insert into TestBfiles values (:1, :2)", [1, bfile]
            )
            self.conn.commit()

    def test_ext_2104(self):
        "E2104 - test reading from a missing file"
        with tempfile.TemporaryDirectory() as temp_dir:
            self._setup_directory(temp_dir)
            file_name = "test_2104.txt"
            self.cursor.execute(
                "select bfilename(:1, :2) from dual",
                [self.dir_name, file_name],
            )
            (bfile,) = self.cursor.fetchone()
            self.assertEqual(bfile.getfilename(), (self.dir_name, file_name))
            self.assertFalse(bfile.fileexists())
            with self.assertRaisesFullCode("ORA-22288"):
                bfile.read()

    def test_ext_2105(self):
        "E2105 - test setting and getting BFILE var"
        with tempfile.TemporaryDirectory() as temp_dir:
            self._setup_directory(temp_dir)
            file_name1 = "test1.txt"
            contents1 = b"extended test 2105 - first file"
            with open(os.path.join(temp_dir, file_name1), "wb") as f:
                f.write(contents1)
            file_name2 = "test2.txt"
            contents2 = b"extended test 2105 - second file"
            with open(os.path.join(temp_dir, file_name2), "wb") as f:
                f.write(contents2)
            var1 = self.cursor.var(oracledb.DB_TYPE_BFILE)
            var2 = self.cursor.var(oracledb.DB_TYPE_BFILE)
            self.cursor.execute(
                f"""
                begin
                    :1 := BFILENAME('{self.dir_name}', '{file_name1}');
                    :2 := BFILENAME('{self.dir_name}', '{file_name2}');
                end;
                """,
                [var1, var2],
            )
            self.assertEqual(var1.getvalue().read(), contents1)
            self.assertEqual(var2.getvalue().read(), contents2)
            self.cursor.execute("begin :1 := :2; end;", [var1, var2])
            self.assertEqual(var1.getvalue().read(), contents2)


if __name__ == "__main__":
    test_env.run_test_cases()



#------------------------------------------------------------------------------
# Copyright (c) 2025, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
# nanoarrow_bridge.pyx
#
# Cython wrapper around the Arrow C Data interface
#------------------------------------------------------------------------------

cimport cpython

from libc.stdint cimport uintptr_t
from libc.string cimport memcpy, strlen, strchr

from .. import errors

cdef extern from "nanoarrow/nanoarrow.c":

    ctypedef int ArrowErrorCode

    ctypedef void (*ArrowBufferDeallocatorCallback)

    cdef struct ArrowBufferAllocator:
        void *private_data

    cdef struct ArrowBuffer:
        uint8_t *data
        int64_t size_bytes
        ArrowBufferAllocator allocator

    cdef union ArrowBufferViewData:
        const void* data

    cdef struct ArrowBufferView:
        ArrowBufferViewData data
        int64_t size_bytes

    cdef struct ArrowArrayView:
        ArrowBufferView *buffer_views

    cdef struct ArrowDecimal:
        pass

    cdef struct ArrowError:
        pass

    cdef struct ArrowStringView:
        const char* data
        int64_t size_bytes

    cdef ArrowErrorCode NANOARROW_OK

    ArrowErrorCode ArrowArrayAllocateChildren(ArrowArray *array,
                                              int64_t n_children)
    ArrowErrorCode ArrowArrayAppendBytes(ArrowArray* array,
                                         ArrowBufferView value)
    ArrowErrorCode ArrowArrayAppendDecimal(ArrowArray* array,
                                           const ArrowDecimal* value)
    ArrowErrorCode ArrowArrayAppendDouble(ArrowArray* array, double value)
    ArrowErrorCode ArrowArrayAppendInt(ArrowArray* array, int64_t value)
    ArrowErrorCode ArrowArrayAppendNull(ArrowArray* array, int64_t n)
    ArrowBuffer* ArrowArrayBuffer(ArrowArray* array, int64_t i)
    ArrowErrorCode ArrowArrayFinishBuildingDefault(ArrowArray* array,
                                                   ArrowError* error)
    ArrowErrorCode ArrowArrayInitFromType(ArrowArray* array,
                                          ArrowType storage_type)
    void ArrowArrayRelease(ArrowArray *array)
    ArrowErrorCode ArrowArrayReserve(ArrowArray* array,
                                     int64_t additional_size_elements)
    ArrowErrorCode ArrowArrayStartAppending(ArrowArray* array)
    ArrowErrorCode ArrowArrayViewInitFromSchema(ArrowArrayView* array_view,
                                                const ArrowSchema* schema,
                                                ArrowError* error)
    ArrowErrorCode ArrowArrayViewSetArray(ArrowArrayView* array_view,
                                          const ArrowArray* array,
                                          ArrowError* error)
    int8_t ArrowBitGet(const uint8_t* bits, int64_t i)
    ArrowBufferAllocator ArrowBufferDeallocator(ArrowBufferDeallocatorCallback,
                                                void *private_data)
    void ArrowDecimalInit(ArrowDecimal* decimal, int32_t bitwidth,
                          int32_t precision, int32_t scale)
    void ArrowDecimalSetBytes(ArrowDecimal *decimal, const uint8_t* value)
    ArrowErrorCode ArrowDecimalSetDigits(ArrowDecimal* decimal,
                                         ArrowStringView value)
    ArrowErrorCode ArrowSchemaDeepCopy(const ArrowSchema *schema,
                                       ArrowSchema *schema_out)
    void ArrowSchemaInit(ArrowSchema* schema)
    ArrowErrorCode ArrowSchemaInitFromType(ArrowSchema* schema, ArrowType type)
    void ArrowSchemaRelease(ArrowSchema *schema)
    ArrowErrorCode ArrowSchemaSetName(ArrowSchema* schema, const char* name)
    ArrowErrorCode ArrowSchemaSetTypeDateTime(ArrowSchema* schema,
                                              ArrowType arrow_type,
                                              ArrowTimeUnit time_unit,
                                              const char* timezone)
    ArrowErrorCode ArrowSchemaSetTypeDecimal(ArrowSchema* schema,
                                             ArrowType type,
                                             int32_t decimal_precision,
                                             int32_t decimal_scale)
    int64_t ArrowSchemaToString(const ArrowSchema* schema, char* out,
                                int64_t n, char recursive)

cdef int _check_nanoarrow(int code) except -1:
    """
    Checks the return code of the nanoarrow function and raises an exception if
    it is not NANOARROW_OK.
    """
    if code != NANOARROW_OK:
        errors._raise_err(errors.ERR_ARROW_C_API_ERROR, code=code)


cdef void pycapsule_array_deleter(object array_capsule) noexcept:
    cdef ArrowArray* array = <ArrowArray*> cpython.PyCapsule_GetPointer(
        array_capsule, "arrow_array"
    )
    if array.release != NULL:
        ArrowArrayRelease(array)
    cpython.PyMem_Free(array)


cdef void pycapsule_schema_deleter(object schema_capsule) noexcept:
    cdef ArrowSchema* schema = <ArrowSchema*> cpython.PyCapsule_GetPointer(
        schema_capsule, "arrow_schema"
    )
    if schema.release != NULL:
        ArrowSchemaRelease(schema)
    cpython.PyMem_Free(schema)


cdef void arrow_buffer_dealloc_callback(ArrowBufferAllocator *allocator,
                                        uint8_t *ptr, int64_t size):
    """
    ArrowBufferDeallocatorCallback for an ArrowBuffer borrowed from
    OracleArrowArray
    """
    cpython.Py_DECREF(<OracleArrowArray> allocator.private_data)


cdef int copy_arrow_array(OracleArrowArray oracle_arrow_array,
                          ArrowArray *src, ArrowArray *dest) except -1:
    """
    Shallow copy source ArrowArray to destination ArrowArray. The source
    ArrowArray belongs to the wrapper OracleArrowArray. The shallow copy idea
    is borrowed from nanoarrow:
    https://github.com/apache/arrow-nanoarrow/main/blob/python
    """
    cdef:
        ArrowBuffer *dest_buffer
        ssize_t i
    _check_nanoarrow(
        ArrowArrayInitFromType(
            dest, NANOARROW_TYPE_UNINITIALIZED
        )
    )

    # Copy metadata
    dest.length = src.length
    dest.offset = src.offset
    dest.null_count = src.null_count

    # Borrow an ArrowBuffer belonging to OracleArrowArray. The ArrowBuffer can
    # belong to an immediate ArrowArray or a child (in case of nested types).
    # Either way, we PY_INCREF(oracle_arrow_array), so that it is not
    # prematurely garbage collected. The corresponding PY_DECREF happens in the
    # ArrowBufferDeAllocator callback.
    for i in range(src.n_buffers):
        if src.buffers[i] != NULL:
            dest_buffer = ArrowArrayBuffer(dest, i)
            dest_buffer.data = <uint8_t *> src.buffers[i]
            dest_buffer.size_bytes = 0
            dest_buffer.allocator = ArrowBufferDeallocator(
                <ArrowBufferDeallocatorCallback> arrow_buffer_dealloc_callback,
                <void *> oracle_arrow_array
            )
            cpython.Py_INCREF(oracle_arrow_array)
        dest.buffers[i] = src.buffers[i]
    dest.n_buffers = src.n_buffers

    # shallow copy of children (recursive call)
    if src.n_children > 0:
        _check_nanoarrow(ArrowArrayAllocateChildren(dest, src.n_children))
        for i in range(src.n_children):
            copy_arrow_array(
                oracle_arrow_array, src.children[i], dest.children[i]
            )


cdef class OracleArrowArray:

    def __cinit__(self, ArrowType arrow_type, str name, int8_t precision,
                  int8_t scale, ArrowTimeUnit time_unit):
        cdef ArrowType storage_type = arrow_type
        self.arrow_type = arrow_type
        self.time_unit = time_unit
        self.name = name
        self.arrow_array = \
                <ArrowArray*> cpython.PyMem_Malloc(sizeof(ArrowArray))
        if arrow_type == NANOARROW_TYPE_TIMESTAMP:
            storage_type = NANOARROW_TYPE_INT64
        if time_unit == NANOARROW_TIME_UNIT_MILLI:
            self.factor = 1e3
        elif time_unit == NANOARROW_TIME_UNIT_MICRO:
            self.factor = 1e6
        elif time_unit == NANOARROW_TIME_UNIT_NANO:
            self.factor = 1e9
        else:
            self.factor = 1

        _check_nanoarrow(ArrowArrayInitFromType(self.arrow_array,
                                                storage_type))
        self.arrow_schema = \
                <ArrowSchema*> cpython.PyMem_Malloc(sizeof(ArrowSchema))
        _check_nanoarrow(ArrowArrayStartAppending(self.arrow_array))
        if arrow_type == NANOARROW_TYPE_DECIMAL128:
            self.precision = precision
            self.scale = scale
            ArrowSchemaInit(self.arrow_schema)
            _check_nanoarrow(ArrowSchemaSetTypeDecimal(self.arrow_schema,
                                                       arrow_type,
                                                       precision, scale))
        else:
            _check_nanoarrow(ArrowSchemaInitFromType(self.arrow_schema,
                                                     storage_type))
            if arrow_type == NANOARROW_TYPE_TIMESTAMP:
                _check_nanoarrow(ArrowSchemaSetTypeDateTime(self.arrow_schema,
                                                            arrow_type,
                                                            time_unit, NULL))
        _check_nanoarrow(ArrowSchemaSetName(self.arrow_schema, name.encode()))

    def __dealloc__(self):
        if self.arrow_array != NULL:
            if self.arrow_array.release != NULL:
                ArrowArrayRelease(self.arrow_array)
            cpython.PyMem_Free(self.arrow_array)
        if self.arrow_schema != NULL:
            if self.arrow_schema.release != NULL:
                ArrowSchemaRelease(self.arrow_schema)
            cpython.PyMem_Free(self.arrow_schema)

    def __len__(self):
        return self.arrow_array.length

    def __repr__(self):
        return (
            f"OracleArrowArray(name={self.name}, "
            f"len={self.arrow_array.length}, "
            f"type={self._schema_to_string()})"
        )

    def __str__(self):
        return self.__repr__()

    cdef str _schema_to_string(self):
        """
        Converts the schema to a string representation.
        """
        cdef char buffer[81]
        ArrowSchemaToString(self.arrow_schema, buffer, sizeof(buffer), 0)
        return buffer.decode()

    cdef int append_bytes(self, void* ptr, int64_t num_bytes) except -1:
        """
        Append a value of type bytes to the array.
        """
        cdef ArrowBufferView data
        data.data.data = ptr
        data.size_bytes = num_bytes
        _check_nanoarrow(ArrowArrayAppendBytes(self.arrow_array, data))

    cdef int append_decimal(self, void* ptr, int64_t num_bytes) except -1:
        """
        Append a value of type ArrowDecimal to the array

        Arrow decimals are fixed-point decimal numbers encoded as a
        scaled integer. decimal128(7, 3) can exactly represent the numbers
        1234.567 and -1234.567 encoded internally as the 128-bit integers
        1234567 and -1234567, respectively.
        """
        cdef:
            ArrowStringView decimal_view
            ArrowDecimal decimal
        decimal_view.data = <char*> ptr
        decimal_view.size_bytes = num_bytes
        ArrowDecimalInit(&decimal, 128, self.precision, self.scale)
        _check_nanoarrow(ArrowDecimalSetDigits(&decimal, decimal_view))
        _check_nanoarrow(ArrowArrayAppendDecimal(self.arrow_array, &decimal))

    cdef int append_double(self, double value) except -1:
        """
        Append a value of type double to the array.
        """
        _check_nanoarrow(ArrowArrayAppendDouble(self.arrow_array, value))

    cdef int append_float(self, float value) except -1:
        """
        Append a value of type float to the array.
        """
        self.append_double(value)

    cdef int append_int64(self, int64_t value) except -1:
        """
        Append a value of type int64_t to the array.
        """
        _check_nanoarrow(ArrowArrayAppendInt(self.arrow_array, value))

    cdef int append_last_value(self, OracleArrowArray array) except -1:
        """
        Appends the last value of the given array to this array.
        """
        cdef:
            int32_t start_offset, end_offset
            ArrowBuffer *offsets_buffer
            ArrowBuffer *data_buffer
            ArrowDecimal decimal
            int64_t *as_int64
            int32_t *as_int32
            double *as_double
            float *as_float
            int8_t as_bool
            int64_t index
            uint8_t *ptr
            void* temp
        if array is None:
            array = self
        index = array.arrow_array.length - 1
        if array.arrow_type in (NANOARROW_TYPE_INT64, NANOARROW_TYPE_TIMESTAMP):
            data_buffer = ArrowArrayBuffer(array.arrow_array, 1)
            as_int64 = <int64_t*> data_buffer.data
            self.append_int64(as_int64[index])
        elif array.arrow_type == NANOARROW_TYPE_DOUBLE:
            data_buffer = ArrowArrayBuffer(array.arrow_array, 1)
            as_double = <double*> data_buffer.data
            self.append_double(as_double[index])
        elif array.arrow_type == NANOARROW_TYPE_FLOAT:
            data_buffer = ArrowArrayBuffer(array.arrow_array, 1)
            as_float = <float*> data_buffer.data
            self.append_double(as_float[index])
        elif array.arrow_type == NANOARROW_TYPE_BOOL:
            data_buffer = ArrowArrayBuffer(array.arrow_array, 1)
            as_bool = ArrowBitGet(data_buffer.data, index)
            self.append_int64(as_bool)
        elif array.arrow_type == NANOARROW_TYPE_DECIMAL128:
            data_buffer = ArrowArrayBuffer(array.arrow_array, 1)
            ArrowDecimalInit(&decimal, 128, self.precision, self.scale)
            ptr = data_buffer.data + index * 16
            ArrowDecimalSetBytes(&decimal, ptr)
            _check_nanoarrow(ArrowArrayAppendDecimal(self.arrow_array,
                                                     &decimal))
        elif array.arrow_type in (
                NANOARROW_TYPE_BINARY,
                NANOARROW_TYPE_STRING
        ):
            offsets_buffer = ArrowArrayBuffer(array.arrow_array, 1)
            data_buffer = ArrowArrayBuffer(array.arrow_array, 2)
            as_int32 = <int32_t*> offsets_buffer.data
            start_offset = as_int32[index]
            end_offset = as_int32[index + 1]
            temp = cpython.PyMem_Malloc(end_offset - start_offset)
            memcpy(temp, &data_buffer.data[start_offset],
                   end_offset - start_offset)
            try:
                self.append_bytes(temp, end_offset - start_offset)
            finally:
                cpython.PyMem_Free(temp)

        elif array.arrow_type in (
                NANOARROW_TYPE_LARGE_BINARY,
                NANOARROW_TYPE_LARGE_STRING
        ):
            offsets_buffer = ArrowArrayBuffer(array.arrow_array, 1)
            data_buffer = ArrowArrayBuffer(array.arrow_array, 2)
            as_int64 = <int64_t*> offsets_buffer.data
            start_offset = as_int64[index]
            end_offset = as_int64[index + 1]
            temp = cpython.PyMem_Malloc(end_offset - start_offset)
            memcpy(temp, &data_buffer.data[start_offset],
                   end_offset - start_offset)
            try:
                self.append_bytes(temp, end_offset - start_offset)
            finally:
                cpython.PyMem_Free(temp)

    cdef int append_null(self) except -1:
        """
        Append a null value to the array.
        """
        _check_nanoarrow(ArrowArrayAppendNull(self.arrow_array, 1))

    cdef int finish_building(self) except -1:
        """
        Finish building the array. No more data will be added to it.
        """
        _check_nanoarrow(ArrowArrayFinishBuildingDefault(self.arrow_array,
                                                         NULL))

    def get_buffer_info(self):
        """
        Get buffer information required by the dataframe interchange logic.
        """
        cdef:
            int64_t n_buffers = self.arrow_array.n_buffers
            ArrowBufferView *buffer
            ArrowArrayView view
        _check_nanoarrow(ArrowArrayViewInitFromSchema(&view, self.arrow_schema,
                                                      NULL))
        _check_nanoarrow(ArrowArrayViewSetArray(&view, self.arrow_array, NULL))

        # initialize all buffers to None to begin with
        buffers = {
            "validity": None,
            "offsets": None,
            "data": None
        }

        # validity buffer
        if n_buffers > 0 and self.arrow_array.null_count > 0:
            buffer = &view.buffer_views[0]
            buffers["validity"] = (
                buffer.size_bytes,
                <uintptr_t> buffer.data.data
            )

        # data / offset buffer
        if n_buffers == 2:
            buffer = &view.buffer_views[1]
            buffers["data"] = (
                buffer.size_bytes,
                <uintptr_t> buffer.data.data
            )
        elif n_buffers == 3:
            buffer = &view.buffer_views[1]
            buffers["offsets"] = (
                buffer.size_bytes,
                <uintptr_t> buffer.data.data
            )
            buffer = &view.buffer_views[2]
            buffers["data"] = (
                buffer.size_bytes,
                <uintptr_t> buffer.data.data
            )

        return buffers

    @property
    def null_count(self) -> int:
        return self.arrow_array.null_count

    @property
    def offset(self) -> int:
        return self.arrow_array.offset

    def __arrow_c_schema__(self):
        """
        Export an ArrowSchema PyCapsule
        """
        cdef ArrowSchema *exported_schema = \
            <ArrowSchema*> cpython.PyMem_Malloc(sizeof(ArrowSchema))
        try:
            _check_nanoarrow(
                ArrowSchemaDeepCopy(
                    self.arrow_schema,
                    exported_schema
                )
            )
        except:
            cpython.PyMem_Free(exported_schema)
            raise
        return cpython.PyCapsule_New(
            exported_schema, 'arrow_schema', &pycapsule_schema_deleter
        )

    def __arrow_c_array__(self, requested_schema=None):
        """
        Returns
        -------
        Tuple[PyCapsule, PyCapsule]
            A pair of PyCapsules containing a C ArrowSchema and ArrowArray,
            respectively.
        """
        if requested_schema is not None:
            raise NotImplementedError("requested_schema")
        cdef ArrowArray *exported_array = \
            <ArrowArray *> cpython.PyMem_Malloc(sizeof(ArrowArray))
        try:
            copy_arrow_array(self, self.arrow_array, exported_array)
            array_capsule = cpython.PyCapsule_New(
                exported_array, 'arrow_array', &pycapsule_array_deleter
            )
        except:
            cpython.PyMem_Free(exported_array)
            raise
        return self.__arrow_c_schema__(), array_capsule
